/* ========================================================================= */
/* === checkA ============================================================== */
/* ========================================================================= */
/* Check that the columns of A are arranged so that the active entries
   are before the inactive entries. Also, check that the active rows and
   free columns of A correspond to AFT. */
/* ========================================================================= */

#include "pproj.h"
void checkA
(
    PPcom     *I,
    int location, /* = 1 if dead rows are still present
                     = 0 otherwise
                     = 2 to check that A' = AT */
    char  *where
)
{
    INT *Ap, *Ai, *Anz, *Ti, *Tp, *AFTp, *AFTnz, *AFTi, *ir, *J,
         i, j, p, q, pp, p1, p2, iprev, nrow, ncol, ni, nmax, tnz ;
    int *ib ;
    double *Ax, *Tx, *AFTx, *X ;
    PPwork    *W ;
    PPprob *Prob ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif
    printf ("check A %s\n", where) ;

    W = I->Work ;
    ir = W->ir ;
    ib = W->ib ;

    Prob = I->Prob ;
    Ap = Prob->Ap ;
    Anz = Prob->Anz ;
    Ai = Prob->Ai ;
    Ax = Prob->Ax ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;
    nmax = MAX (nrow, ncol) ;
    if ( location == 2 )
    {
        INT *ATp, *ATi, *Work ;
        double *ATx ;
        ATp = (INT *) malloc ((nrow+1)*sizeof (INT)) ;
        ATi = (INT *) malloc (Ap [ncol]*sizeof (INT)) ;
        Work= (INT *) malloc (nrow*sizeof (INT)) ;
        ATx = (double *) malloc (Ap [ncol]*sizeof (double)) ;
        pproj_transpose (ATp, ATi, ATx, Ap, Ai, Ax, nrow, ncol, Work) ;
        for (i = 0; i <= nrow; i++)
        {
            if ( ATp [i] != W->ATp [i] )
            {
                printf ("ATp [%i] = %i != W->ATp [%i] = %i\n",
                         i, ATp [i], i, W->ATp [i]) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
        for (i = 0; i < Ap [ncol]; i++)
        {
            if ( ATi [i] != W->ATi [i] )
            {
                printf ("ATi [%i] = %i != W->ATi [%i] = %i\n",
                         i, ATi [i], i, W->ATi [i]) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
        for (i = 0; i < Ap [ncol]; i++)
        {
            if ( ATx [i] != W->ATx [i] )
            {
                printf ("ATx [%i] = %e != W->ATx [%i] = %e\n",
                         i, ATx [i], i, W->ATx [i]) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
        free (ATp) ;
        free (ATi) ;
        free (ATx) ;
        free (Work) ;
        return ;
    }

    AFTp = W->AFTp ;
    AFTi = W->AFTi ;
    AFTnz = W->AFTnz ;
    AFTx = W->AFTx ;

    for (i = 0 ; i <= nrow ; i++) ASSERT (AFTp [i] == W->ATp [i]) ;

    /* allocate workspace */
    J = malloc (nmax*sizeof (INT)) ;
    X = malloc (nmax*sizeof (double)) ;

    /* check for sorted columns */
    for (j = 0 ; j < ncol ; j++)
    {
        if ( Anz [j] < 0 )
        {
            printf ("Anz [%i] = %i < 0\n", j, Anz [j]) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        p1 = Ap [j] ;
        p2 = Ap [j] + Anz [j] ;
        iprev = -1 ;
        for (p = p1; p < p2; p++)
        {
            i = Ai [p] ;
            if (i >= nrow || i < 0)
            {
                printf("i: %i j: %i nrow: %i ncol: %i\n", i,j,nrow,ncol);
                pproj_error (-1, __FILE__, __LINE__, where);
            }
            if (i <= iprev)
            {
                printf("col: %i (entries in column not increasing)\n", j) ;
                for (p = p1 ; p < p2 ; p++)
                {
                    printf ("p: %i i: %i\n", p, Ai [p]) ;
                }
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
            iprev = i ;
            /* if ( (ir [i] > ni) || (ir [i] < -ni) )
            {
                printf ("col: %i has active row %i, "
                        "but it is dropped ir: %i\n", j, i, ir [i]) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            } */
        }
        for (; p < Ap [j+1]; p++)
        {
            i = Ai [p] ;
            if ( ir [i] <= ni )
            {
                printf ("col: %i row: %i inactive, but ir: %i\n",
                         j, i, ir [i]);
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
    }

/* check that each entry in A is the transpose of AFT */
/* compute transpose of A */
   
    for (i = 0; i < nrow; i++)
    {
        J [i] = 0 ;
    }

    for (j = 0 ; j < ncol ; j++)
    {
        if ( !ib [j] )
        {
            p1 = Ap [j] + Anz [j] ;
            for (p = Ap [j]; p < p1; p++)
            {
                if ( ir [Ai [p]] <= ni ) J [Ai [p]]++ ;
            }
        }
    }

    Tp = malloc ((nrow+1)*sizeof (int)) ;
    Tp [0] = 0 ;
    for (i = 0 ; i < nrow ; i++)
    {
        Tp [i+1] = Tp [i] + J [i] ;
        J [i] = Tp [i] ;
    }
    tnz = Tp [nrow] ; 
    if (tnz == 0)
    {
	/* nothing to do; free workspace and return */
	free (J) ;
	free (X) ;
	free (Tp) ;
        return ;
    }

    Ti = malloc (tnz*sizeof (int)) ;
    Tx = malloc (tnz*sizeof (double)) ;

    /* ====================================================================== */
    /* === T = A' =========================================================== */
    /* ====================================================================== */

    for (j = 0 ; j < ncol ; j++)
    {
        if ( !ib [j] )
        {
            p1 = Ap [j] + Anz [j] ;
            for (p = Ap [j] ; p < p1 ; p++)
            {
                if ( ir [Ai [p]] <= ni )
                {
                    pp = J [Ai [p]]++ ;
                    Ti [pp] = j ;
                    Tx [pp] = Ax [p] ;
                }
            }
        }
    }

/* check that each row of T agree with the corresponding row of AFT */

    for (j = 0; j < ncol; j++) J [j] = 0 ;
    for (i = 0; i < nrow; i++)
    {
        if ( ir [i] <= ni )
        {
            p1 = Tp [i+1] ;
            if ( AFTnz [i] != p1 - Tp [i] )
            {
                printf ("row: %i AFTnz (%i) != Tp (%i+1) - Tp (%i)\n",
                    i, AFTnz [i], p1, Tp [i]) ;
                for (p = AFTp [i]; p < AFTp [i] + AFTnz [i]; p++)
                {
                    printf ("AFT matrix: row: %i col: %i x: %e\n",
                        i, AFTi [p], AFTx [p]) ;
                }
                for (p = Tp [i]; p < p1; p++)
                {
                    printf ("A matrix, row: %i col: %i x: %e\n",
                        i, Ti [p], Tx [p]) ;
                }
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
            for (p = Tp [i]; p < p1; p++)
            {
                j = Ti [p] ;
                J [j] = 1 ;
                X [j] = Tx [p] ;
            }
            p = AFTp [i] ;
            p1 = p + AFTnz [i] ;
            for (; p < p1; p++)
            {
                j = AFTi [p] ;
                if ( !J [j] )
                {
                    printf ("row: %i col: %i in AFT, but not in A transpose\n",
                        i, j) ;
                    pproj_error (-1, __FILE__, __LINE__, where) ;
                }
                J [j] = 0 ;
                if ( AFTx [p] != X [j] )
                {
                    printf ("row: %i col: %i AFTx(%e) != Ax (%e)\n",
                        i, j, AFTx [p], X [j]) ;
                    pproj_error (-1, __FILE__, __LINE__, where) ;
                }
            }
        }
    }

    /* In dasa check that all rows in the active part of A are active rows.
       In one location in the code, cannot make this test since dead
       rows are still present. */
    if ( location == 0 )
    {
        for (j = 0; j < ncol; j++)
        {
            q = Ap [j] + Anz [j] ;
            for (p = Ap [j]; p < q; p++)
            {
                if ( ir [Ai [p]] > ni ) /* row is dropped */
                {
                    printf ("row: %i col: %i in active part of A, but "
                            "row is dropped since ir = %i > ni = %i\n",
                             Ai [p], j, ir [Ai [p]], ni) ;
                    pproj_error (-1, __FILE__, __LINE__, where) ;
                }
            }
        }
    }

    /* free workspace */
    free (J) ;
    free (X) ;
    free (Tp) ;
    free (Ti) ;
    free (Tx) ;
}
