/* ========================================================================= */
/* === check_AFT =========================================================== */
/* ========================================================================= */

#include "pproj.h"

void check_AFT
(
    PPcom    *I,
    int  use_ir,  /* TRUE means skip deleted rows */
    char *where
)
{
    INT *ATp, *ATi, *AFTp, *AFTi, *AFTnz, *J, *ib, *ir,
        i, j, p, q, nrow, ncol, ni ;
    double *X, *ATx, *AFTx, *y ;
    PPprob *Prob ;
    PPwork    *W ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif
    printf ("check AFT %s\n", where) ;

    Prob = I->Prob ;
    y = Prob->y ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;

    W = I->Work ;
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;
    AFTp = W->AFTp ;
    AFTi = W->AFTi ;
    AFTx = W->AFTx ;
    AFTnz = W->AFTnz ;
    ib = W->ib ;
    ir = W->ir ;

    for (i = 0 ; i <= nrow ; i++) ASSERT (AFTp [i] == ATp [i]) ;

    /* allocate workspace */
    J = (INT *) malloc (ncol*sizeof (INT)) ;
    X = (double *) malloc (ncol*sizeof (double)) ;

    for (j = 0; j < ncol; j++) J [j] = 0 ;
    for (i = 0; i < nrow; i++)
    {
        /* skip deleted rows if use_ir is TRUE and the row has been deleted */
        if ( use_ir && (ir [i] <= ni) || !use_ir )
        {
            p = AFTp [i] ;
            q = p + AFTnz [i] ;
            for (; p < q; p++)
            {
                j = AFTi [p];
                if ( (j < 0) || (j > ncol) )
                {
                    PRINTF ("AFT error %s, col: %i in AFT out of range "
                            "(%i to %i)\n", where, j, 0, ncol-1) ;
                    pproj_error (-1, __FILE__, __LINE__, "stop") ;
                }
                if ( J [j] )
                {
                    PRINTF ("AFT error %s, row: %i col: "
                            "%i repeats in AFT\n", where, i, j) ;
                    pproj_error (-1, __FILE__, __LINE__, "stop") ;
                }
                if ( ib [j] )
                {
                    PRINTF ("AFT error %s, row: %i col: %i in AFT "
                            "but ib not zero: %i\n", where, i, j, ib [j]) ;
                    pproj_error (-1, __FILE__, __LINE__, "stop") ;
                }
                J [j] = 1 ;
                X [j] = AFTx [p] ;
            }
            p = ATp [i] ;
            q = ATp [i+1] ;
            for (; p < q; p++)
            {
                j = ATi [p];
                if ( (j < 0) || (j >= ncol) )
                {
                    PRINTF ("AFT error %s, col: %i in AT out of range "
                            "(%i to %i)\n", where, j, 0, ncol-1) ;
                    pproj_error (-1, __FILE__, __LINE__, "stop") ;
                }
                if ( !ib [j] )
                {
                    if ( !J [j] )
                    {
                        PRINTF ("AFT error %s, row: %i col: %i in AT and "
                                "ib zero, but not in AFT\n", where, i, j) ;
                        pproj_error (-1, __FILE__, __LINE__, "stop") ;
                    }
                    J [j] = 0 ;
                    if ( ATx [p] != X [j] )
                    {
                        PRINTF ("AFT error %s, row: %i col: %i "
                                "val in  AT (%e) != AFT (%e)\n",
                                 where, i, j, ATx [p], X [j]) ;
                        pproj_error (-1, __FILE__, __LINE__, "stop") ;
                    }
                }
            }
            p = AFTp [i] ;
            q = p + AFTnz [i] ;
            for (; p < q; p++)
            {
                j = AFTi [p];
                if ( J [j] )
                {
                    PRINTF ("AFT error %s, row: %i col: %i "
                            "in AFT but not in AT\n", where, i, j) ;
                    pproj_error (-1, __FILE__, __LINE__, "stop") ;
                }
                J [j] = 0 ;
            }
        }
    }

    /* free workspace */
    free (J) ;
    free (X) ;
}
