/* ========================================================================= */
/* === checkdiag3 ========================================================== */
/* ========================================================================= */
/* check the diagonal of LDL' between toprow and botrow */
/* ========================================================================= */

#include "pproj.h"
void check_diag3
(
    PPcom   *I,
    INT toprow,
    INT botrow,
    int   chol  /* T => the matrix was factorized, otherwise updates */
)
{
    INT i, p, q, nrow, ni, *ATp, *ATi, *Lp, *Li, *Lnz, *ir ;
    int *ib ;
    double *ATx, *Lx, *D, *E, t, err, sigma, di ;
    cholmod_factor *L ;
    PPwork *W ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif

    W = I->Work ;
    ir = W->ir ;
    ib = W->ib ;
    sigma = I->Work->sigma ;

    ATp   = W->ATp ;
    ATi   = W->ATi ;
    ATx   = W->ATx ;
    nrow = I->Prob->nrow ;
    ni = I->Prob->ni ;

    L = W->L ;
    Lp   = L->p ;
    Lnz  = L->nz ;
    Li   = L->i ;
    Lx   = L->x ;

    /* allocate workspace */
    E = malloc (nrow*sizeof (double)) ;
    D = malloc (nrow*sizeof (double)) ;

    pproj_initx (D, ZERO, nrow) ;
    pproj_initx (E, ZERO, nrow) ;
    err = 0 ;
    for (i = toprow; i < botrow; i++)
    {
        if ( ir [i] <= ni )
        {
            t = sigma ;
            q = ATp [i+1] ;
            for (p = ATp [i]; p < q; p++)
            {
                if ( ib [ATi [p]] == 0 )
                {
                    t += ATx [p]*ATx [p] ;
                }
            }
            p = Lp [i] ;
	    di = Lx [p] ;   	/* D(i,i) */
            D [i] = t ;
            E [i] += di ;
            err = MAX (err, fabs (t - E [i])) ;
            q = p + Lnz [i] ;
            t = di ;
            for (p++ ; p < q ; p++)
            {
                E [Li [p]] += t * Lx [p] * Lx [p] ;
            }
        }
    }

    if ( I->Parm->PrintLevel > 1 )
        printf ("diag3, err in factorization: %e chol: %i\n", err, chol);
    if ( chol )
    {
        if ( err/nrow > 1.e-11 )
        {
            printf ("err in chol: %e\n", err);
            for (i = toprow; i < botrow;i++)
            {
                if ( ir [i] < 0 )
                {
                    printf ("i: %i D: %e E: %e err: %e\n",
                        i, D [i], E [i], fabs (D [i] - E [i])) ;
                }
            }
            pproj_error (-1, __FILE__, __LINE__, "partial chol failure") ;
        }
    }
    else
    {
        if ( err/nrow > 1.e-3 )
        {
            printf ("err in chol update: %e\n", err);
            for (i = toprow; i < botrow;i++)
            {
                if ( ir [i] <= ni )
                {
                    printf ("i: %i D: %e E: %e err: %e\n",
                        i, D [i], E [i], fabs (D [i] - E [i])) ;
                }
            }
            pproj_error (-1, __FILE__, __LINE__, "partial chol failure") ;
        }
    }

    /* free workspace */
    free (E) ;
    free (D) ;
}
