/* ========================================================================== */
/* check_dual (based on the current ib, only enforce bounds when ib [j] != 0) */
/* ========================================================================== */

#include "pproj.h"
double check_dual /* return dual objective at current point */
(
    PPcom    *I,
    double  *dl,
    char *where,
    int    save, /* TRUE  => saves dual objective value for next comparison
                    FALSE => put dual value in C->mark but do not compare */
    int   check  /* TRUE => check for increase in dual objective */
)
{
    INT    ni, i, j, k, p, q, ncol, nrow, *ATi, *ATp, *ib, *ir, *ineq_row ;
    double lmax, zmax, cost, t, s, zerr, tol,
          *ATx, *bl, *bu, *hi, *lo, *y, *z, *lambda ;
    PPcheck   *C ;
    PPprob *Prob ;
    PPwork    *W ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif
    printf ("check dual value %s\n", where) ;

    C = I->Check ;
    Prob = I->Prob ;
    y = Prob->y ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    hi = Prob->hi ;
    lo = Prob->lo ;
    bl = Prob->bl ;
    bu = Prob->bu ;
    ni = Prob->ni ;
    ineq_row = Prob->ineq_row ;

    W = I->Work ;
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;
    ib  = W->ib ;
    ir  = W->ir ;

    /* allocate workspace */
    z = malloc (ncol*sizeof (double)) ;
    lambda = malloc (nrow*sizeof (double)) ;

    if ( dl == NULL )
    {
        for (i = 0; i < nrow; i++)
        {
            lambda [i] = W->dlambda [i] + W->lambda [i] + W->shift_l [i] ;
        }
    }
    else
    {
        pproj_copyx (lambda, dl, nrow) ;
        for (i = 0; i < nrow; i++)
        {
            lambda [i] += W->dlambda [i] + W->lambda [i] + W->shift_l [i] ;
        }
    }
    pproj_copyx (z, y, ncol) ;
    p = 0 ;
    lmax = ZERO ;
    s = ZERO ;
    /* compute z = y + A'*lambda and lambda - shift_l */
    for (i = 0; i < nrow; i++)
    {
        t = lambda [i] ;
        if ( fabs (t) > lmax ) lmax = fabs (t) ;
        q = ATp [i+1] ;
        for (; p < q; p++)
        {
            z [ATi [p]] += t*ATx [p] ;
        }
        t -= W->shift_l [i] ;
        s += t*t ;
    }
    /* the proximal part of the objective */
    cost = -0.5*I->Work->sigma*s ;
    C->lmax = lmax ;

    /* check sign of lambda for inequalities */
    tol = I->Parm->checktol*lmax ;
    for (i = 0; i < nrow; i++)
    {
        if ( ir [i] == 0 )
        {
            continue ;
        }
        if ( ir [i] < 0 ) /* should have lambda_i >= 0 */
        {
            if ( lambda [i] < -tol )
            {
                printf ("i: %i ir: %i lambda: %e\n", i, ir [i], lambda [i]) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
        else if ( ir [i] > ni ) /* should have lambda_i = 0 */
        {
            if ( fabs (lambda [i]) > tol )
            {
                printf ("i: %i ir: %i lambda: %e\n", i, ir [i], lambda [i]) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
        else /* should have lambda_i <= 0 */
        {
            if ( lambda [i] > tol )
            {
                printf ("i: %i ir: %i lambda: %e\n", i, ir [i], lambda [i]) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
    }
    zmax = ZERO ;
    s = ZERO ;
    for (j = 0; j < ncol; j++)
    {
        if ( ib [j] )
        {
            if      ( z [j] < lo [j] ) z [j] = lo [j] ;
            else if ( z [j] > hi [j] ) z [j] = hi [j] ;
        }
        if ( fabs (z [j]) > zmax ) zmax = fabs (z [j]) ;
        s += (z [j] - y [j])*(z [j] - y [j]) ;
    }
    cost += 0.5*s ;
    if ( zmax < ONE )
    {
        zmax = ONE ;
    }
    C->zmax = zmax ;
    for (j = 0; j < ncol; j++)
    {
        zerr = ZERO ;
        if      ( ib [j] < 0 ) zerr = fabs (z [j] - lo [j]) ;
        else if ( ib [j] > 0 ) zerr = fabs (z [j] - hi [j]) ;
        if ( zerr != ZERO )
        {
            if ( zmax != ZERO ) zerr /= zmax ;
            if ( zerr > I->Parm->checktol )
            {
                printf ("j: %i z: %e lo: %e hi: %e ib: %i\n",
                         j, z [j], lo [j], hi [j], ib [j]) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
    }

    k = 1 ;
    for (i = 0; i < nrow; i++)
    {
        if ( ir [i] == 0 )
        {
            t = Prob->b [i] ;
        }
        else if ( ir [i] < 0 )
        {
            t = bl [k] ;
            k++ ;
        }
        else if ( ir [i] <= ni )
        {
            t = bu [k] ;
            k++ ;
        }
        else
        {
            k++ ;
            continue ;
        }
        q = ATp [i+1] ;
        for (p = ATp [i]; p < q; p++)
        {
            t -= ATx [p]*z [ATi [p]] ;
        }
        cost += t*lambda [i] ;
    }
    if ( I->Parm->PrintLevel > 0 ) printf("VAL: %25.15e\n",cost) ;
    if ( check == TRUE )
    {
        if ( C->prior > cost + I->Parm->checktol*(C->zmax+C->lmax) )
        {
            PRINTF ("new cost %e  < old cost %e\n", cost, C->prior) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }
    if ( save == TRUE )
    {
        C->prior = cost ;
    }
    else
    {
        C->mark = cost ;
    }

    /* free workspace */
    free (z) ;
    free (lambda) ;
    return (cost) ;
}
