/* ========================================================================== */
/* check_dualsimple (evaluate dual at the current point ignoring ib and ir) */
/* ========================================================================== */

#include "pproj.h"
double check_dualsimple /* return dual objective at current point */
(
    PPcom    *I,
    double  *dl
)
{
    INT    ni, i, j, k, p, q, ncol, nrow, *ATi, *ATp, *ib, *ir, *ineq_row ;
    double lmax, cost, t, s,
          *ATx, *bl, *bu, *hi, *lo, *y, *z, *lambda ;
    PPprob *Prob ;
    PPwork    *W ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif

    Prob = I->Prob ;
    y = Prob->y ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    hi = Prob->hi ;
    lo = Prob->lo ;
    bl = Prob->bl ;
    bu = Prob->bu ;
    ni = Prob->ni ;
    ineq_row = Prob->ineq_row ;

    W = I->Work ;
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;
    ib  = W->ib ;
    ir  = W->ir ;

    /* allocate workspace */
    z = malloc (ncol*sizeof (double)) ;
    lambda = malloc (nrow*sizeof (double)) ;

    if ( dl == NULL )
    {
        for (i = 0; i < nrow; i++)
        {
            lambda [i] = W->dlambda [i] + W->lambda [i] + W->shift_l [i] ;
        }
    }
    else
    {
        pproj_copyx (lambda, dl, nrow) ;
        for (i = 0; i < nrow; i++)
        {
            lambda [i] += W->dlambda [i] + W->lambda [i] + W->shift_l [i] ;
        }
    }
    pproj_copyx (z, y, ncol) ;
    p = 0 ;
    lmax = ZERO ;
    s = ZERO ;
    /* compute z = y + A'*lambda and lambda - shift_l */
    for (i = 0; i < nrow; i++)
    {
        t = lambda [i] ;
        if ( fabs (t) > lmax ) lmax = fabs (t) ;
        q = ATp [i+1] ;
        for (; p < q; p++)
        {
            z [ATi [p]] += t*ATx [p] ;
        }
        t -= W->shift_l [i] ;
        s += t*t ;
    }
    /* the proximal part of the objective */
    cost = -0.5*I->Work->sigma*s ;
    s = ZERO ;
    for (j = 0; j < ncol; j++)
    {
        if      ( z [j] < lo [j] ) z [j] = lo [j] ;
        else if ( z [j] > hi [j] ) z [j] = hi [j] ;
        s += (z [j] - y [j])*(z [j] - y [j]) ;
    }
    cost += 0.5*s ;

    k = 1 ;
    for (i = 0; i < nrow; i++)
    {
        if ( ir [i] == 0 )
        {
            t = Prob->b [i] ;
        }
        else if ( lambda [i] < 0 )
        {
            t = bu [k] ;
            k++ ;
        }
        else if ( lambda [i] > 0 )
        {
            t = bl [k] ;
            k++ ;
        }
        else
        {
            k++ ;
            continue ;
        }
        q = ATp [i+1] ;
        for (p = ATp [i]; p < q; p++)
        {
            t -= ATx [p]*z [ATi [p]] ;
        }
        cost += t*lambda [i] ;
    }

    /* free workspace */
    free (z) ;
    free (lambda) ;
    return (cost) ;
}
