/* ========================================================================= */
/* === check_forward  ====================================================== */
/* ========================================================================= */

#include "pproj.h"
#define ISNAN(x) ((x) != (x))
void check_forward
(
    PPcom        *I,
    double *forward,
    double       *r,
    int    *joblist,
    int          nj,
    char     *where
)
{
    INT *row_start, *leftdesc, *Ap, *Anz, *Ai, *Lp, *Li, *Lnz, *ir,
         botblk, topblk, toprow, botrow, i, j, p, q, nrow, ncol, ni, jobnum ;
    int *ib ;
    double *Ax, *Lx, *rhs, *cold, *dlambda, mx, mxr, t ;
    cholmod_factor *L ;
    PPwork         *W ;
    PPprob      *Prob ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif

    printf ("check forward solve %s\n", where) ;
    if ( nj == 0 ) return ;
    botrow = 0 ;
    W = I->Work ;
    ib = W->ib ;
    ir = W->ir ;
    cold = W->cold ;
    dlambda = W->dlambda ;

    Prob = I->Prob ;
    Ap   = Prob->Ap ;
    Anz  = Prob->Anz ;
    Ai   = Prob->Ai ;
    Ax   = Prob->Ax ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni   = Prob->ni ;

    L = W->L ;
    Lp   = L->p ;
    Lnz  = L->nz ;
    Li   = L->i ;
    Lx   = L->x ;

    leftdesc = W->leftdesc ;
    row_start = W->row_start ;

    /* allocate workspace */
    rhs = malloc (nrow*sizeof (double)) ;

    for (i = 0; i < nrow; i++) rhs [i] = r [i] ;
    mx = ONE ;
    mxr = ZERO ;
    for (j = 0; j < ncol; j++)
    {
        if ( ib [j] == 0 )
        {
            t = cold [j] ;
            mx = MAX (mx, fabs (t)) ;
            q = Ap [j] + Anz [j] ;
            for (p = Ap [j]; p < q; p++)
            {
                rhs [Ai [p]] -= Ax [p]*t ;
            }
        }
    }

    i = 0 ;
    for (jobnum = 0; jobnum < nj; jobnum++)
    {
        botblk = joblist [jobnum] ;
        botrow = row_start [botblk+1] ;
        topblk = leftdesc [botblk] ;  /* blks in range topblk:botblk */
        toprow = row_start [topblk] ;
        for (; i < toprow; i++) rhs [i] = 0 ;
        for (i = toprow; i < botrow; i++)
        {
            if ( ir [i] <= ni )
            {
                mxr = MAX (mxr, fabs (rhs [i])) ;
                mx = MAX (mx, fabs (forward [i])) ;
                mx = MAX (mx, fabs (r [i])) ;
                mx = MAX (mx, fabs (dlambda [i])) ;
                t  = forward [i] ;
                rhs [i] -= t ;
                p = Lp [i] ;
                q = p + Lnz [i] ;
                for (p++ ; p < q ; p++)
                {
                    rhs [Li [p]] -= Lx [p]*t ;
                }
            }
        }
    }

    for (i = botrow; i < nrow; i++) rhs [i] = ZERO ;
    t = ZERO ;
    for (i = 0; i < nrow; i++)
    {
        if ( ir [i] <= ni )
        {
            t = MAX (t, fabs (rhs [i])) ;
        }
    }

    mx = MAX (mx, mxr) ;
    if ( I->Parm->PrintLevel > 1 )
        printf ("err in forward solve: %e mx: %e\n", t, mx);
    if ( mx != ZERO  || ISNAN (mx) || ISNAN (t))
    {
        t /= mx ;
        if ( t > 1.e-6 || ISNAN (mx) || ISNAN (t))
        {
            printf ("err in forward solve: %e mx: %e\n", t, mx);
            for (jobnum = 0; jobnum < nj; jobnum++)
            {
                botblk = joblist [jobnum] ;
                botrow = row_start [botblk+1] ;
                topblk = leftdesc [botblk] ;  /* blks in range topblk:botblk */
                toprow = row_start [topblk] ;
                for (i = toprow; i < botrow; i++)
                {
                    if ( ir [i] <= ni ) printf("i: %i rhs: %e\n", i , rhs [i]) ;
                }
            }
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }

    /* free workspace */
    free (rhs) ;
}
