/* ========================================================================= */
/* === check_link ========================================================== */
/* ========================================================================= */

#include "pproj.h"
void check_link
(
    PPcom     *I,
    int *joblist,
    int       nj,
    char  *where
)
{
    INT ncol, nrow, nf, ni, ni1, ni2, ntot, i, i0, i1, j, j0, j1, k,
        botblk, botrow, botcol, topblk, toprow, topcol, row,
        *RLinkUp, *RLinkDn, *F, *ir, *lLinkUp, *lLinkDn, *uLinkUp, *uLinkDn,
        *lstart, *ustart, *ineq_row, *Rstart, *row_start, *sol_start,
        *leftdesc, *temp ;
    int blks, jobnum, *ib, *ib_old ;
    PPprob *Prob ;
    PPwork    *W ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif
    printf ("check links %s\n", where) ;

    Prob = I->Prob ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;
    ni1= ni + 1 ;
    ni2= ni + 2 ;
    ntot = ni + ncol + 1 ;

    W = I->Work ;
    ib_old = W->ib ;
    ir = W->ir ;

    /* allocate workspace */
    F    = malloc (ncol*sizeof (int)) ;
    ib   = malloc (ncol*sizeof (int)) ;
    temp = malloc (MAX (ntot, nrow)*sizeof (int)) ;

    nf = 0 ;
    for (j = 0; j < ncol; j++)
    {
        ib [j] = ib_old [j] ;
        if ( ib [j] == 0 ) F [nf++] = j ;
    }

    RLinkUp = W->RLinkUp ;
    RLinkDn = W->RLinkDn ;

    uLinkUp = W->SLinkUp ;
    uLinkDn = W->SLinkDn ;
    lLinkUp = W->SLinkUp ;
    lLinkDn = W->SLinkDn ;

    for (j = 0; j < ncol; j++) temp [j] = 0 ;

    for (k = 0; k < nf; k++)
    {
        j = F [k] ;
        if ( (j < 0) || (j >= ncol) )
        {
            printf ("F: %i, out of range 0 to %i, at k: %i\n", j, ncol, k) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( temp [j] == 1 )
        {
            printf ("F: %i, appears twice at k: %i\n", j, k) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        temp [j] = 1 ;
    }

    for (i = 0; i < nrow; i++) temp [i] = 0 ;
    i0 = nrow ;

    while ( (i1 = RLinkUp [i0]) < nrow )
    {
        if ( i1 < 0 )
        {
            printf ("RLinkUp: %i less than zero at %i\n", i1, i0) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( ir [i1] > ni )
        {
            printf ("row: %i in RLinks but ir = %i > ni = %i\n", i1,ir[i1],ni) ;
            pproj_error (-1, __FILE__, __LINE__, "stop") ;
        }
        if ( RLinkDn [i1] != i0 )
        {
            printf ("RLinkDn: %i, at %i does not match Up at %i\n",
                RLinkDn [i1], i1, i0) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( temp [i1] != 0 )
        {
            printf ("%i in RLinkUp appears twice in link\n", i1) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        temp [i1] = 1 ;
        i0 = i1 ;
    }
    for (i = 0; i < nrow; i++)
    {
        if ( ir [i] <= ni )
        {
           if ( !temp [i] )
           {
               printf ("ir: %i <= ni = %i but row: %i not in RLinks\n",
                        ir [i], ni, i) ;
               pproj_error (-1, __FILE__, __LINE__, "stop") ;
           }
        }
    }
    if ( nj == 0 )
    {
        free (temp) ;
        free (ib) ;
        free (F) ;
        return ;
    }

/* check block structure */

    blks = W->blks ;
    leftdesc = W->leftdesc ;
    sol_start = W->sol_start ;
    lstart = W->lstart ;
    ustart = W->ustart ;
    Rstart = W->Rstart ;
    row_start = W->row_start ;
    ineq_row = Prob->ineq_row ;

    for (jobnum = 0; jobnum < nj; jobnum++)
    {
        botblk = joblist [jobnum] ;
        botrow = row_start [botblk+1] ;
        topblk = leftdesc [botblk] ;  /* blks in range topblk:botblk */
        toprow = row_start [topblk] ;
        for (i = toprow; i < botrow; i++) temp [i] = 0 ;
        i = Rstart [botblk] ;

        while ( i < botrow )
        {
            if ( i < toprow )
            {
                printf ("job: %i botblk: %i row: %i < toprow: %i\n",
                    jobnum, botblk, i, toprow) ;
                pproj_error (-1, __FILE__, __LINE__, "stop") ;
            }
            if ( temp [i] == 1 )
            {
                printf ("job: %i botblk: %i row: %i repeats\n",
                    jobnum, botblk, i) ;
                pproj_error (-1, __FILE__, __LINE__, "stop") ;
            }
            temp [i] = 1 ;
            if ( ir [i] > ni ) /* row was dropped */
            {
                printf ("job: %i botblk: %i row: %i ir: %i, row in links "
                        "but inactive\n", jobnum, botblk, i, ir [i]) ;
                pproj_error (-1, __FILE__, __LINE__, "stop") ;
            }
            i = RLinkUp [i] ;
        }
        for (i = toprow; i < botrow; i++)
        {
            if ( ir [i] <= ni )
            {
                if ( !temp [i] ) /* active rows should be in linked list above*/
                {
                    printf ("job: %i botblk: %i row: %i ir: %i, "
                       "missing from links\n", jobnum, botblk, i, ir [i]) ;
                    pproj_error (-1, __FILE__, __LINE__, "stop") ;
                }
            }
        }
    }

    for (jobnum = 0; jobnum < nj; jobnum++)
    {
        botblk = joblist [jobnum] ;
        botcol = sol_start [botblk+1] ; /* first inequality row of next block */
        botrow = row_start [botblk+1] ;
        topblk = leftdesc [botblk] ;  /* blks in range topblk:botblk */
        topcol = sol_start [topblk] ; /* first inequality row */
        toprow = row_start [botblk] ;

        for (j = topcol; j < botcol; j++) temp [j] = 0 ;
        j = ustart [botblk] ;

        while ( j < botcol )
        {
            if ( j < topcol )
            {
                printf ("check ustart, sol\n") ;
                printf ("job: %i botblk: %i col: %i < topcol: %i\n",
                    jobnum, botblk, j, topcol) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
            if ( temp [j] == 1 )
            {
                printf ("job: %i botblk: %i col: %i repeats\n",
                    jobnum, botblk, j) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
            i = ineq_row [j] ;
            if ( ir [i] != j )
            {
                printf ("job: %i botblk: %i col: %i i: %i ir: %i, ni: %i\n",
                          jobnum, botblk, j, i, ir [i], ni) ;
                printf ("column in ulinks but not at upper bound check\n") ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
            temp [j] = 1 ;
            j = uLinkUp [j] ;
        }
    }

    for (jobnum = 0; jobnum < nj; jobnum++)
    {
        botblk = joblist [jobnum] ;
        botcol = sol_start [botblk+1] ;
        botrow = row_start [botblk+1] ;
        topblk = leftdesc [botblk] ;  /* blks in range topblk:botblk */
        topcol = sol_start [topblk] ;
        toprow = row_start [botblk] ;
        for (j = topcol; j < botcol; j++) temp [j] = 0 ;
        j = lstart [botblk] ;

        while ( j < botcol )
        {
            if ( j < topcol )
            {
                printf ("check lstart, sol\n") ;
                printf ("job: %i botblk: %i col: %i < topcol: %i\n",
                    jobnum, botblk, j, topcol) ;
                pproj_error (-1, __FILE__, __LINE__, "stop") ;
            }
            if ( temp [j] == 1 )
            {
                printf ("job: %i botblk: %i col: %i repeats\n",
                    jobnum, botblk, j) ;
                pproj_error (-1, __FILE__, __LINE__, "stop") ;
            }
            i = ineq_row [j] ;
            if ( ir [i] != -j )
            {
                printf ("job: %i botblk: %i col: %i i: %i ir: %i, col in "
                        "llinks but ir not = %i\n",
                         jobnum, botblk, j, i, ir [i], -j) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
            temp [j] = 1 ;
            j = lLinkUp [j] ;
        }
    }

    for (j = lLinkUp [ni1]; j <= ni; j = lLinkUp [j])
    {
        row = ineq_row [j] ;
        if ( ir [row] != -j )
        {
            printf("row: %i ir: %i row in lLink (j = %i) when dropped\n",
                row, ir [row], j) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }
    for (j = uLinkUp [ni2]; j <= ni; j = uLinkUp [j])
    {
        row = ineq_row [j] ;
        if ( ir [row] != j )
        {
            printf("row: %i ir: %i row in uLink (j = %i) when dropped\n",
                row, ir [row], j) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }

    for (j = 0; j <= ni; j++) temp [j] = ZERO ;
    j0 = ni1 ;
    while ( (j1 = lLinkUp [j0]) <= ni )
    {
        if ( (j1 < 1 ) || (j1 > ni) )
        {
            printf ("lLinkUp: %i, out of range %i to %i at %i\n",
                j1, 1, ni, j0) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( lLinkDn [j1] != j0 )
        {
            printf ("lLinkDn: %i, at %i does not match Up at %i\n",
                lLinkDn [j1], j1, j0) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( temp [j1] != 0 )
        {
            printf ("%i in lLinkUp appears twice in link\n", j1) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        temp [j1] = 1 ;
        j0 = j1 ;
    }

    for (j = 0; j <= ni; j++) temp [j] = ZERO ;
    j0 = ni2 ;
    while ( (j1 = uLinkUp [j0]) <= ni )
    {
        if ( (j1 < 1 ) || (j1 > ni) )
        {
            printf ("uLinkUp: %i, out of range %i to %i at %i\n",
                j1, 1, ni, j0) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( uLinkDn [j1] != j0 )
        {
            printf ("uLinkDn: %i, at %i does not match Up at %i\n",
                uLinkDn [j1], j1, j0) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( temp [j1] != 0 )
        {
            printf ("%i in uLinkUp appears twice in link\n", j1) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        temp [j1] = 1 ;
        j0 = j1 ;
    }

/*  printf ("Links OK at %s\n", where) ; */

    /* free workspace */
    free (F) ;
    free (ib) ;
    free (temp) ;
}
