/* ========================================================================= */
/* === check_modlist ======================================================= */
/* ========================================================================= */

#include "pproj.h"
void check_modlist
(
    PPcom    *I,
    char *where
)
{
    INT *RowmodList, *ColmodList, *RowmodFlag, *ColmodFlag,
        *ColFlag, *RowFlag, *ir, *ineq_row,
         nrow, ncol, nrowadd, nrowdel, ncoladd, ncoldel, i, j, k, ni ;
    int *ib ;
    PPprob *Prob ;
    PPwork    *W ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif
    printf ("check mod list %s\n", where) ;

    Prob = I->Prob ;
    nrow = Prob->nrow ;
    ncol = Prob->ncol ;
    ni = Prob->ni ;
    ineq_row = Prob->ineq_row ;
    W = I->Work ;
    nrowadd = W->nrowadd ;
    nrowdel = W->nrowdel ;
    ncoladd = W->ncoladd ;
    ncoldel = W->ncoldel ;
    ir = W->ir ;
    ib = W->ib ;

    if ( nrowadd < 0 )
    {
        PRINTF ("nrowadd (%i) < 0 at %s\n", nrowadd, where) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }
    if ( nrowdel < 0 )
    {
        PRINTF ("nrowdel (%i) < 0 at %s\n", nrowdel, where) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }
    if ( ncoladd < 0 )
    {
        PRINTF ("ncoladd (%i) < 0 at %s\n", ncoladd, where) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }
    if ( ncoldel < 0 )
    {
        PRINTF ("ncoldel (%i) < 0 at %s\n", ncoldel, where) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }

    if ( nrowadd >= nrow )
    {
        PRINTF ("nrowadd (%i) >= nrow (%i) at %s\n", nrowadd, nrow, where) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }
    if ( nrowdel >= nrow )
    {
        PRINTF ("nrowdel (%i) >= nrow (%i) at %s\n", nrowdel, nrow, where) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }
    if ( ncoladd >= ncol )
    {
        PRINTF ("ncoladd (%i) >= ncol (%i) at %s\n", ncoladd, ncol, where) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }
    if ( ncoldel >= ncol )
    {
        PRINTF ("ncoldel (%i) >= ncol (%i) at %s\n", ncoldel, ncol, where) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }

    RowmodList = W->RowmodList ;
    RowmodFlag = W->RowmodFlag ;
    ColmodList = W->ColmodList ;
    ColmodFlag = W->ColmodFlag ;

/* check range of list */

    for (i = 1; i <= nrowadd; i++)
    {
        if ( RowmodList [ni-i] <= 0 )
        {
            PRINTF ("%i. Add RowmodList [%i] = %i < 0 at %s\n",
                i, nrow-i, RowmodList [nrow-i], where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( RowmodList [ni-i] > ni )
        {
            PRINTF ("%i. Add RowmodList [%i] = %i >= nrow (%i) at %s\n",
                i, nrow-i, RowmodList [nrow-i], nrow, where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }
    for (i = 0; i < nrowdel; i++)
    {
        if ( RowmodList [i] <= 0 )
        {
            PRINTF ("Del RowmodList [%i] = %i < 0 at %s\n",
                i, RowmodList [i], where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( RowmodList [i] > ni )
        {
            PRINTF ("Del RowmodList [%i] = %i >= nrow (%i) at %s\n",
                i, RowmodList [i], nrow, where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }

    for (i = 1; i <= ncoldel; i++)
    {
        if ( ColmodList [ncol-i] < 0 )
        {
            PRINTF ("%i. Del ColmodList [%i] = %i < 0 in %s\n",
                i, ncol-i, ColmodList [i], where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( ColmodList [ncol-i] >= ncol )
        {
            PRINTF ("%i. Del ColmodList [%i] = %i >= ncol (%i) in %s\n",
                i, ncol-i, ColmodList [i], ncol, where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }
    for (i = 0; i < ncoladd; i++)
    {
        if ( ColmodList [i] < 0 )
        {
            PRINTF ("Add ColmodList [%i] = %i < 0 at %s\n",
                i, ColmodList [i], where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        if ( ColmodList [i] >= ncol )
        {
            PRINTF ("Add ColmodList [%i] = %i >= ncol (%i) at %s\n",
                i, ColmodList [i], ncol, where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }

/*  look for repetitions in list */
    ColFlag = (INT *) malloc (ncol*sizeof (INT)) ;
    RowFlag = (INT *) malloc ((ni+1)*sizeof (INT)) ;

    for (i = 0; i <= ni; i++) RowFlag [i] = EMPTY ;
    for (j = 0; j < ncol; j++) ColFlag [j] = EMPTY ;

    for (i = 1; i <= nrowadd; i++)
    {
        k = RowmodList [ni-i] ;
        if ( RowFlag [k] != EMPTY )
        {
            PRINTF ("Add RowmodList [%i] = %i repeats at %s\n", i, k, where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        RowFlag [k] = 1 ;
    }

    for (i = 0; i < nrowdel; i++)
    {
        k = RowmodList [i] ;
        if ( RowFlag [k] != EMPTY )
        {
            PRINTF ("Del RowmodList [%i] = %i repeats at %s\n",
                nrow-i, k, where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        RowFlag [k] = 1 ;
    }

    for (i = 1; i <= ncoldel; i++)
    {
        k = ColmodList [ncol-i] ;
        if ( ColFlag [k] != EMPTY )
        {
            PRINTF ("Del ColmodList [%i] = %i repeats at %s\n", i, k, where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        ColFlag [k] = 1 ;
    }

    for (i = 0; i < ncoladd; i++)
    {
        k = ColmodList [i] ;
        if ( ColFlag [k] != EMPTY )
        {
            PRINTF ("Add ColmodList [%i] = %i repeats at %s\n",
                ncol-i, k, where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
        ColFlag [k] = 1 ;
    }

/* check that the modflags are only nonempty when there is a row or column
   in modlist */

    for (i = 0; i <= ni; i++) RowFlag [i] = EMPTY ;
    for (i = 0; i < nrowdel; i++) RowFlag [RowmodList [i]] = i ;
    for (i = 1; i <= nrowadd; i++) RowFlag [RowmodList [ni-i]] = ni - i ;

    for (i = 1; i <= ni; i++)
    {
        if ( RowmodFlag [i] != EMPTY )
        {
            /* is the row in RowmodList? */
            if ( RowFlag [i] == EMPTY )
            {
                printf ("RowmodFlag: %i nrowadd: %i nrowdel: %i ni: %i\n",
                         RowmodFlag [i], nrowadd, nrowdel, ni) ;
                printf ("RowmodList: %i\n", RowmodList [RowmodFlag [i]]) ;
                PRINTF ("ineq %i has flag set, but not in modlist (%s)\n",
                    i, where) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
            /* does the RowmodFlag point to the correct row? */
            if ( RowFlag [i] != RowmodFlag [i] )
            {
                PRINTF("ineq %i, RowmodFlag points to row %i (incorrect), %s\n",
                         i, RowmodList [RowmodFlag [i]], where) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
        else
        {
            if ( RowFlag [i] != EMPTY )
            {
                PRINTF ("ineq %i is in the modList, but flag not set, %s\n",
                i, where) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
    }

    for (j = 0; j < ncol; j++) ColFlag [j] = EMPTY ;
    for (j = 0; j < ncoladd; j++) ColFlag [ColmodList [j]] = j ;
    for (j = 1; j <= ncoldel; j++) ColFlag [ColmodList [ncol-j]] = ncol - j ;

    for (j = 0; j < ncol; j++)
    {
        if ( ColmodFlag [j] != EMPTY )
        {
            /* is the col in ColmodList? */
            if ( ColFlag [j] == EMPTY )
            {
                PRINTF ("col %i has flag set, but not in modlist, %s\n",
                    j, where) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
            /* does the ColmodFlag point to the correct col? */
            if ( ColFlag [j] != ColmodFlag [j] )
            {
                PRINTF ("col %i, ColmodFlag points to col %i (incorrect), %s\n",
                         j, ColmodList [ColmodFlag [j]], where) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
        else
        {
            if ( ColFlag [j] != EMPTY )
            {
                PRINTF ("col %i is in the modList, but flag not set, %s\n",
                    j, where) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
    }

    /* free workspace */
    free (ColFlag) ;
    free (RowFlag) ;

/* check sign of ir and ib for compatibility with deletions and additions */
    for (i = 1; i <= nrowadd; i++)
    {
        k = ineq_row [RowmodList [ni-i]] ;
        if ( ir [k] > ni )
        {
            PRINTF ("%i. Add RowmodList [%i] = %i while ir = %i >= 0 in %s\n",
            i, nrow-i, k, ir [k], where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }

    for (i = 0; i < nrowdel; i++)
    {
        k = ineq_row [RowmodList [i]] ;
        if ( ir [k] <= ni )
        {
            PRINTF ("Del RowmodList [%i] = %i while ir = %i < 0 at %s\n",
            i, k, ir [k], where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }

    for (i = 1; i <= ncoldel; i++)
    {
        k = ColmodList [ncol-i] ;
        if ( ib [k] == 0 )
        {
            PRINTF ("%i. Del ColmodList [%i] = %i while ib = %i at %s\n",
            i, ncol-i, k, ib [k], where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }

    for (i = 0; i < ncoladd; i++)
    {
        k = ColmodList [i] ;
        if ( ib [k] != 0 )
        {
            PRINTF ("Add ColmodList [%i] = %i while ib = %i != 0 at %s\n",
            ncol-i, k, ib [k], where) ;
            pproj_error (-1, __FILE__, __LINE__, where) ;
        }
    }
}
