/* ========================================================================== */
/* ==== checkb ============================================================== */
/* ========================================================================== */

#include "pproj.h"
void checkb
(
    PPcom    *I,
    char *where
)
{
    int location ;
    INT    *Ai, *Ap, *ib, *ir, *ineq_row, *temp, *Heap,
            ni, i, ineqindex, j, p, q, ncol, nrow, row ;
    double zmax, t, err, bmax,
           *x, *Ax, *b, *bl, *bu, *hi, *lo, *lambda ;
    PPcheck   *C ;
    PPprob *Prob ;
    PPwork    *W ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif

    printf ("checkb %s\n", where) ;
    C = I->Check ;
    location = C->location ;
    Prob = I->Prob ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    Ap = Prob->Ap ;
    Ai = Prob->Ai ;
    Ax = Prob->Ax ;
    bl = Prob->bl ;
    bu = Prob->bu ;
    ni = Prob->ni ;
    ineq_row = Prob->ineq_row ;

    W = I->Work ;
    ib = W->ib ;
    ir = W->ir ;
    lo = W->lo ;
    hi = W->hi ;
    lambda = W->lambda ;
    x = W->x ;
    b = (double *) malloc (nrow*sizeof (double)) ;
    lambda = (double *) malloc (nrow*sizeof (double)) ;
    pproj_copyx (b, Prob->b, nrow) ;
    bmax = ZERO ;
    for (i = 0; i < nrow; i++)
    {
        lambda [i] = W->dlambda [i] + W->lambda [i] + W->shift_l [i] ;
        if ( fabs (b [i]) > bmax )
        {
            bmax = fabs (b [i]) ;
        }
    }
    if ( location == 4 ) /* in ssor1, ir fudged, fix it */
    {
        temp = (INT *) malloc (nrow*sizeof (INT)) ;
        pproj_copyI (temp, ir, nrow) ;
        i = ncol + ni ;
        Heap = W->arrayi ;
        while ( Heap [i] != EMPTY )
        {
            ineqindex = Heap [i] ;
            row = ineq_row [ineqindex] ;
            temp [row] = ni + ineqindex ;
            i-- ;
        }
        ir = temp ;
    }

    for (i = 1; i <= ni; i++)
    {
        row = ineq_row [i] ;
        if ( ir [row] < 0 )
        {
            b [row] = bl [i] ;
            if ( fabs (bl [i]) > bmax )
            {
                bmax = fabs (bl [i]) ;
            }
            if ( (ir [row] != -i) || (lambda [row] < ZERO) )
            {
                printf ("checkb error %s, row: %i ir: %i should be %i lambda"
                        ": %e\n", where, row, ir [row], -i, lambda [row]) ;
                pproj_error (-1, __FILE__, __LINE__, "stop") ;
            }
        }
        else if ( (ir [row] > 0) && (ir [row] <= ni) )
        {
            b [row] = bu [i] ;
            if ( fabs (bu [i]) > bmax )
            {
                bmax = fabs (bu [i]) ;
            }
            if ( (ir [row] != i) || (lambda [row] > ZERO) )
            {
                printf ("checkb error %s, row: %i ir: %i should be %i lambda"
                        ": %e\n", where, row, ir [row], i, lambda [row]) ;
                pproj_error (-1, __FILE__, __LINE__, "stop") ;
            }
        }
        else
        {
            if ( ir [row] == 0 )
            {
                printf ("checkb error %s, equation %i is inequality "
                        "but ir = 0", where, row) ;
                pproj_error (-1, __FILE__, __LINE__, "stop") ;
            }
            b [row] = ZERO ;
        }
    }
    for (j = 0; j < ncol; j++)
    {
        t = x [j] ;
        if ( t != ZERO )
        {
            q = Ap [j+1] ;
            for (p = Ap [j]; p < q; p++)
            {
                b [Ai [p]] -= t*Ax [p] ;
            }
        }
    }

    /* In sparsa, x is not updated in each iteration, but b includes current
       bound variables. Need to return to ib to adjust b. */
    if ( location == 6 )
    {
        for (j = 0; j < ncol; j++)
        {
            if ( ib [j] > 0 )
            {
                t = hi [j] ;
            }
            else if ( ib [j] < 0 )
            {
                t = lo [j] ;
            }
            else
            {
                t = ZERO ;
            }
            if ( t != ZERO )
            {
                q = Ap [j+1] ;
                for (p = Ap [j]; p < q; p++)
                {
                    b [Ai [p]] -= t*Ax [p] ;
                }
            }
        }
    }

    /* for all location except location 1 (phase 1) we also subtract
       sigma*lambda from b */
    if ( location != 1 )
    {
        pproj_saxpy (b, W->lambda, -I->Work->sigma, nrow) ;
    }
    bmax = MAX (bmax, ONE) ;
    zmax = C->zmax ;
    if ( zmax <= ONE ) zmax = ONE ;
    err = ZERO ;
    for (i = 0; i < nrow; i++)
    {
        t = fabs (b [i] - W->b [i])/(zmax+bmax+C->lmax) ;
        if ( t > I->Parm->checktol )
        {
            printf ("checkb error %s, row: %i b: %e should be: %e\n",
                     where, i, W->b [i], b [i]) ;
            pproj_error (-1, __FILE__, __LINE__, "stop") ;
        }
        if ( err < t ) err = t ;
    }
    if ( I->Parm->PrintLevel > 0 ) PRINTF("b error: %e\n", err) ;
    free (b) ;
    free (lambda) ;
    if ( location == 4 )
    {
        free (temp) ;
    }
}
