/* ========================================================================== */
/* ==== check_error in c ==================================================== */
/* ========================================================================== */

#include "pproj.h"
void checkc
(
    PPcom    *I,
    char *where
)
{
    INT *ATp, *ATi, *ir, i, j, p, nrow, ncol, ni ;
    double *ATx, *c, t, err, norm_l ;
    PPprob *Prob ;
    PPwork    *W ;

#ifdef NDEBUG
    fprintf (stderr, "Debugging is on!\n") ; abort () ;
#endif

    printf ("checkc %s\n", where) ;
    Prob = I->Prob ;
    ncol = Prob->ncol ;
    nrow = Prob->nrow ;
    ni = Prob->ni ;
    W = I->Work ;
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;
    ir = W->ir ;

    /* allocate workspace */
    c = (double *) malloc (ncol*sizeof (double)) ;
    pproj_copyx (c, Prob->y, ncol) ;
    if ( I->Check->location != 1 ) /* outside of phase1, x subtracted from c */
    {
        pproj_saxpy (c, W->x, -ONE, ncol) ;
    }
    norm_l = ZERO ;
    for (i = 0; i < nrow; i++)
    {
        t = W->dlambda [i] + W->lambda [i] + W->shift_l [i] ;
        if ( ir [i] > ni ) /* the row has been dropped */
        {
            if ( t != 0 )
            {
                printf ("row %i has ir = %i "
                        "and lambda = %e != 0\n", i, ir [i], t) ;
                pproj_error (-1, __FILE__, __LINE__, where) ;
            }
        }
        else
        {
            norm_l += fabs (t) ;
            for (p = ATp [i]; p < ATp [i+1]; p++)
            {
                c [ATi [p]] += ATx [p]*t ; /* c = y + A'lambda */
            }
        }
    }

    if ( norm_l == ZERO )
    {
        norm_l = 1. ;
    }

    err = ZERO ;
    for (j = 0; j < ncol; j++)
    {
/*printf ("j: %i cerr: %e\n", j, fabs(c [j] - W->c [j])) ;*/
        err = MAX (err, fabs (c [j] - W->c [j])) ;
    }

    if ( err/norm_l > I->Parm->checktol )
    {
        printf ("relative err in c: %e norm_l: %e\n", err, norm_l) ;
        pproj_error (-1, __FILE__, __LINE__, where) ;
    }

    /* free workspace */
    free (c) ;
}
