/* ========================================================================== */
/* ==== pproj_initlevels ==================================================== */
/* ========================================================================== */

#include "pproj.h"

int pproj_initlevels
(
    PPcom *I
)
{
    INT       *rowperm, *colperm, *invperm, *rowcount, *ineq_row, *sortindices,
              *Ai, *Ap, *Anz, *ATp, *ATi, *AFTp,
              *col_start, *row_start, *blk_preperm,
              *parent, *ncols, *row_to_blk, *sol_to_blk, *newrows,
              *twork, *tworkstart,
               Annz, blk, maxdepth, maxcol, maxrow, i, j, k, ki, kd, l, m,
               n, p, pp, q, iend, col, root, blks, print, nrow, ncol, ni, ntot ;

    int    *Cmember, *Cparent, *depth, *nkids, *Kids, *Kp, *leftdesc ;

    double    Amax, tic, t, *Ax, *ATx, *b, *bl, *bu, *tempb, *lo, *hi,
              *workb, *worky, *worklo, *workhi ;

    PPparm *Parm ;
    PPprob *Prob ;
    PPstat *Stat ;
    PPwork    *W ;
    cholmod_common *cmm ;
#ifndef NDEBUG
    /* check that matrix is correctly permuted */
    double *CheckAx, *CheckATx, *OrigAx ;
    INT *CheckAi, *CheckAp, *CheckATi, *CheckATp, *OrigAi, *OrigAp ;
#endif
    W = I->Work ;
    cmm = W->cmm ;

    Parm = I->Parm ;
    print = Parm->PrintLevel ;
    Stat = I->Stat ;
    Prob = I->Prob ;    /* the problem structure */
    ncol = Prob->ncol ; /* number of columns in A */
    nrow = Prob->nrow ; /* number of rows in A */

    if ( Parm->providePPcom )
    {
        /* copy Prob->b, Prob->lo, and Prob->hi to work structure */
        pproj_copyx (W->b, Prob->b, nrow) ;
        pproj_copyx (W->lo, Prob->lo, ncol) ;
        pproj_copyx (W->hi, Prob->hi, ncol) ;
        if ( Parm->cholmod )
        {
            /* -------------------------------------------------------------- */
            /* if use_lambda, permute lambda by rowperm and save in shift_l */
            /* -------------------------------------------------------------- */
            if ( Parm->use_lambda )
            {
                rowperm = Prob->rowperm ;
                for (i = 0; i < nrow; i++)
                {
                    W->shift_l [i] = W->lambda [rowperm [i]] ;
                }
                pproj_initx (W->lambda, ZERO, nrow) ;
            }
            else /* lambda = shift_l = 0 */
            {
                pproj_initx (W->shift_l, ZERO, nrow) ;
                pproj_initx (W->lambda, ZERO, nrow) ;
            }
            maxdepth = Stat->maxdepth ;
            /* Stat->solves counts number of solves at each level in the tree */
            for (k = 0; k <= maxdepth; k++)
            {
                Stat->solves [k] = 0 ;
            }
        }
        else
        {
            if ( Parm->use_lambda )
            {
                pproj_copyx (W->shift_l, W->lambda, nrow) ;
                pproj_initx (W->lambda, ZERO, nrow) ;
            }
            else
            {
                pproj_initx (W->shift_l, ZERO, nrow) ;
            }
        }
        return (PPROJ_STATUS_OK) ;
    }
    ni   = Prob->ni ;   /* number of strict inequalities */
    Ax   = Prob->Ax ;   /* numerical entries in A */
    Ai   = Prob->Ai ;   /* row indices in A */
    Ap   = Prob->Ap ;   /* column pointers for A */
    Anz  = Prob->Anz ;  /* number of nonzeros in each column of A */
    Annz = Ap [ncol] ;  /* maximum number of nonzeros in A */

    /* rowperm [i] is the row number of the original matrix corresponding
       to row i of the permuted matrix */
    rowperm = Prob->rowperm ;

    /* colperm [j] = column of original matrix corresponding to column j in
                     permuted matrix */
    colperm = Prob->colperm ;

/* AT matrix was allocated in initwork, and assigned entries below */
    ATp = W->ATp ;
    ATi = W->ATi ;
    ATx = W->ATx ;

/* The pointers to the start of rows for the AFT matrix are computed below */
    AFTp = W->AFTp ;

    /* ---------------------------------------------------------------------- */
    /* fill-reducing permutation */
    /* ---------------------------------------------------------------------- */

    /* The rows of the original matrix A will be permuted to reduce fill-in.
       rowperm [i] is the row number of the original matrix that
       is inserted in row i of the permuted matrix. In a multilevel
       partitioning, Cmember [i] is the block number associated with row i
       in the original matrix. Cparent is an array of size equal to the
       number of blocks, and Cparent [k] is the parent of node (block) k in
       the tree associated with the multilevel partition. Think of the
       permuted matrix having the rows associated with block 0 at the
       top of the matrix, followed by the rows associated with block 1,
       and so on. The trailing rows in the matrix correspond to the
       root of the tree. */
       
    if ( Parm->cholmod )
    {
#ifndef NDEBUG
    OrigAp = (INT *) malloc ((ncol+1)*sizeof (INT)) ;
    OrigAi = (INT *) malloc (Annz*sizeof (INT)) ;
    OrigAx = (double *) malloc (Annz*sizeof (double)) ;
    /* save a copy of the original matrix */
    pproj_copyI (OrigAp, Ap, ncol+1) ;
    pproj_copyI (OrigAi, Ai, Annz) ;
    pproj_copyx (OrigAx, Ax, Annz) ;
#endif

        tic = pproj_timer ( ) ;
        Cmember = malloc (nrow*sizeof (INT)) ; /* freed later */
        Cparent = malloc (nrow*sizeof (INT)) ; /* freed later */
    
        if ( Parm->multilevel ) /* perform a multilevel partition of A */
        {
            double nd_oksep = 0.06 ; /* was 0.12 */
            int nd_small = 500 ;     /* was 500 */
    
            /* -------------------------------------------------------------- */
            /* multilevel case:  partition and order the matrix */
            /* -------------------------------------------------------------- */
    
            /* compute maximum number of nonzeros in a column if it was
               not previously computed */
            if ( Prob->maxcol < 0 )
            {
                maxcol = 0 ;
                p = 0 ;
                for (j = 1; j <= ncol; j++)
                {
                    q = Ap [j] ;
                    if ( q - p > maxcol ) maxcol = q - p ;
                    p = q ;
                }
                Prob->maxcol = maxcol ;
            }
    
            if (Prob->maxcol > 0.5 * nrow)
            {
                if ( print )
                {
                    printf ("-- ordering method: colamd (forcing one block)\n");
                }
                cholmod_colamd (W->A, NULL, 0, TRUE, rowperm, cmm) ;
                blks = 1 ;
            }
            else
            {
    
                cmm->current = 0 ;
    
                /* use old default method - order each connected component
                 * separately.  This can result in an unbalanced separator tree.
                 */
                cmm->method [0].nd_components = 1 ;
    
                /* use new default method
                cmm->method [0].nd_components = 0 ;
                */
    
                if ( print )
                {
                    printf ("nd_components: %d\n",
                             cmm->method [0].nd_components) ;
                }
    
                blks = cholmod_nested_dissection (W->A, NULL, 0, rowperm,
                        Cparent, Cmember, cmm) ;
    
                if ( print )
                {
                    printf ("orig blks %d\n", blks) ;
                }
    
#ifndef NDEBUG
                for (i = 0 ; i < nrow ; i++)
                {
                    k = Cmember [i] ;
                    if (k < 0 || k > blks) printf ("Hey orig! %d %d\n", i, k) ;
                    ASSERT (k >= 0 && k < blks) ;
                }
#endif
    
                /* prune the separator tree */
                if ( print )
                {
                    printf ("prune %g %d\n", nd_oksep, nd_small) ;
                }
                blks = cholmod_collapse_septree (nrow, blks, nd_oksep, nd_small,
                       Cparent, Cmember, cmm) ;
            }
    
            if ( print )
            {
                printf ("blks: %d\n", blks) ;
            }
    
            /* Fix Cparent.  CHOLMOD can return a forest, but PPROJ expects a
             * tree.  Force all root nodes < blks-1 to have node blks-1 as their
             * parent. */
    
            for (k = 0 ; k < blks-1 ; k++)
            {
                if (Cparent [k] == EMPTY)
                {
                    Cparent [k] = blks-1 ;
                }
            }
    
        }
        else /* not multilevel, reorder rows to reduce fill-in */
        {
            blks = 1 ; /* there is only one block */
            /* -------------------------------------------------------------- */
            /* CHOLMOD nested dissection, or COLAMD if dense cols exist: */
            /* -------------------------------------------------------------- */
    
            if ( print )
            {
                printf ("--- ordering method: nd or colamd\n") ;
                printf ("maxcol %d  nrow %d   maxcol/nrow %8.5f  "
                        "maxcol/sqrt(nrow) %8.5f\n",
                Prob->maxcol, nrow, Prob->maxcol/((double) nrow),
                Prob->maxcol/sqrt((double) nrow)) ;
            }
    
            if (Prob->maxcol > 0.5 * nrow)
            {
    
                if ( print )
                {
                    printf ("--- ordering method: colamd\n") ;
                }
                cholmod_colamd (W->A, NULL, 0, TRUE, rowperm, cmm) ;
            }
            else
            {
                /* use old default method - order each connected component
                 * separately.  This can result in an unbalanced separator
                 * tree. */
                cmm->current = 0 ;
                cmm->method [0].nd_components = 1 ;
    
                if ( print )
                {
                    printf ("--- ordering method: nd\n") ;
                }
                cholmod_nested_dissection (W->A, NULL, 0, rowperm,
                    Cparent, Cmember, cmm) ;
            }
        }
    
        if ( print )
        {
            printf ("\n--------------- pproj_initlevels has found rowperm\n") ;
        }

        Stat->partition = pproj_timer ( ) - tic ;
    
        /* ------------------------------------------------------------------ */
        /* allocate remainder of the multilevel part of W */
        /* ------------------------------------------------------------------ */
        /* the tree describing the multilevel decomposition is quite small,
           int or short is fine */
        W->Kids      = (int *) malloc (blks*sizeof (INT)) ;
        W->nkids     = (int *) malloc (blks*sizeof (INT)) ;
        W->parent    = (int *) malloc (blks*sizeof (INT)) ;
        W->Kp        = (int *) malloc (blks*sizeof (INT)) ;
        W->depth     = (int *) malloc (blks*sizeof (INT)) ;
        W->leftdesc  = (int *) malloc (blks*sizeof (int)) ;
    
        /* The following relate to the matrix where INT could be long */
        W->col_start = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        W->sol_start = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        W->row_start = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        W->col_to_blk= (int *) malloc (ncol*sizeof (INT)) ;
        W->sol_to_blk= (int *) malloc ((ni+2)*sizeof (INT)) ;

        /* The following arrays change, and are stored in the work structure */
        W->joblist = (int *) malloc (blks*sizeof (int)) ;
        W->kidsleft = (int *) malloc (blks*sizeof (int)) ;
        W->jobcols = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        W->jobrows = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        W->Rstart = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        W->Rend = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        if ( W->Rend == NULL )
        {
            return (PPROJ_OUT_OF_MEMORY) ;
        }
    
        row_start = W->row_start ;
        if ( blks > 1 )
        {
            /* compute row_start.  After permuting the matrix, block k starts at
             * row row_start [k] and ends at row row_start [k+1]-1. */
            for (k = 0 ; k < blks ; k++)
            {
                row_start [k] = 0 ;
            }
            for (i = 0 ; i < nrow ; i++)
            {
                k = Cmember [i] ;
#ifndef NDEBUG
                if (k < 0 || k > blks) printf ("Hey! %d %d\n", i, k) ;
                ASSERT (k >= 0 && k < blks) ;
#endif
                row_start [k]++ ;
            }
            q = 0 ;
            for (k = 0 ; k < blks ; k++)
            {
                p = q ;
                q += row_start [k] ;
                row_start [k] = p ;
            }
            row_start [blks] = nrow ;
        }
        else /* single level, 1 block */
        {
            Cparent [0] = EMPTY ;
            row_start [0] = 0 ;
            row_start [1] = nrow ;
        }
        free (Cmember) ;
    
        nkids = W->nkids ;
        leftdesc = W->leftdesc ;
        col_start = W->col_start ;
        depth = W->depth ;
    
        /*
        Given:
            row_start [blk] = first row in blk
            parent    [blk] = parent of node blk
    
        Compute:
            leftdesc  [blk] = left descendant of blk
            nkid      [blk] = number of children of blk
            col_start [blk] = first column in blk
            sol_start [blk] = index of first singleton in blk
            col_to_blk[j]   = k if column j (after colperm) is in block k
            sol_to_blk[j]   = k if jth strict inequality is in block k
    
        Work array:
            ncols      [blk] = number of columns in blk
            blk_preperm[j]   = blk associated with column j before the perm
            row_to_blk [i]   = k if row i after rowperm is in block k 
    
        */
    
        b = Prob->b ;
        workb = W->b ;
        /* ------------------------------------------------------------------ */
        /* permute b according to rowperm */
        /* ------------------------------------------------------------------ */
        for (i = 0; i < nrow; i++)
        {
            ASSERT (rowperm [i] >= 0 && rowperm [i] < nrow) ;
            workb [i] = b [rowperm [i]] ;
        }
        pproj_copyx (b, workb, nrow) ;
    
        /* ------------------------------------------------------------------ */
        /* if use_lambda, permute lambda by rowperm and save in shift_l */
        /* ------------------------------------------------------------------ */
        if ( Parm->use_lambda )
        {
            for (i = 0; i < nrow; i++)
            {
                W->shift_l [i] = W->lambda [rowperm [i]] ;
            }
            pproj_initx (W->lambda, ZERO, nrow) ;
        }
        else /* lambda = shift_l = 0 */
        {
            pproj_initx (W->shift_l, ZERO, nrow) ;
            pproj_initx (W->lambda, ZERO, nrow) ;
        }

        hi = Prob->hi ;
        lo = Prob->lo ;
        ineq_row = Prob->ineq_row ;
    
        /* Temporary work space that will be freed below */
        tworkstart = twork = (INT *) malloc ((ncol+3*nrow+blks)*sizeof (INT)) ;
        if ( twork == NULL )
        {
            return (PPROJ_OUT_OF_MEMORY) ;
        }
        invperm     = twork ; twork += nrow ;
        rowcount    = twork ; twork += nrow ;
        blk_preperm = twork ; twork += ncol ;
        ncols       = twork ; twork += blks ;
        row_to_blk  = twork ; twork += nrow ;
#ifndef NDEBUG
        invperm = (INT *) malloc (nrow*sizeof (INT)) ;
        rowcount = (INT *) malloc (nrow*sizeof (INT)) ;
        blk_preperm = (INT *) malloc (ncol*sizeof (INT)) ;
        ncols = (INT *) malloc (blks*sizeof (INT)) ;
        row_to_blk = (INT *) malloc (nrow*sizeof (INT)) ;
#endif
    
        root = blks - 1 ;
        for (k = 0; k < blks; k++)
        {
            ncols [k] = nkids [k] = 0 ;
            depth [k] = leftdesc [k] = EMPTY ;
        }
        depth [root] = 0 ;
    
    /* Evaluate number of children and left descendant */
    
        leftdesc [root] = 0 ;
        i = 0 ;
        maxdepth = 0 ;
        parent = W->parent ;
        for (k = 0; k < root; k++)
        {
            iend = row_start [k+1] ;
            for (; i < iend; i++)
            {
                row_to_blk [i] = k ;
            }
    
            parent [k] = Cparent [k] ;
            nkids [Cparent [k]]++ ;
            if ( nkids [k] == 0 ) leftdesc [k] = k ;
            /* go up the tree and set the left descendants of the successive
               parents to be k if they have not yet been set. This loop
               terminates since the left descendant of the root has been
               initialized to 0 */
            l = k ;
            while ( leftdesc [(l = Cparent [l])] == EMPTY )
            {
                leftdesc [l] = k ;
            }
            l = k ;
            m = 0 ;
            while ( depth [l] == EMPTY )
            {
                l = Cparent [l] ;
                m++ ;
            }
            m += depth [l] ;
            maxdepth = MAX (maxdepth, m) ;
            l = k ;
            while ( depth [l] == EMPTY )
            {
                depth [l] = m-- ;
                l = Cparent [l] ;
            }
        }
        for ( ; i < nrow ; i++)
        {
            row_to_blk [i] = root ;
        }
        Stat->maxdepth = maxdepth ;
        /* Stat->solves counts number of solves at each level in the tree. */
        Stat->solves = (INT *) malloc ((maxdepth+1)*sizeof (INT)) ;
    
        /* initialize the solve statistics */
        for (k = 0; k <= maxdepth; k++)
        {
            Stat->solves [k] = 0 ;
        }
    
        ASSERT (row_start [root+1] == nrow) ;
    
    /* Invert rowperm, initialize row counters */
        for (i = 0; i < nrow; i++)
        {
            /* invperm [i] = row in permuted matrix corresponding to row i
                             in original matrix */
            invperm [rowperm [i]] = i ;
            rowcount [i] = 0 ;
        }
        /* Transform ineq_row indices to row indices in the permuted matrix */
        newrows = ATp ;/* need nrow space, ATp has not been used yet */
        for (i = 0; i < ni; i++)
        {
            /*ineq_row was shifted by 1*/
            newrows [i] = invperm [ineq_row [i+1]] ;
        }
        /* sort newrows indices in increasing order */
        sortindices = AFTp ; /* AFTp has not been used yet */
        pproj_iminsort (sortindices, newrows, ineq_row /*work space*/, ni) ;
        for (i = 0; i < ni; i++)
        {
            /* ineq_row continues to be shifted by 1 */
            ineq_row [i+1] = newrows [sortindices [i]] ;
        }
        /* Also apply the permutation given in sortindices to bl and bu.
           W->D is allocated arrays not yet used */
        tempb = W->D ; 
        bl = Prob->bl ;
        for (i = 0; i < ni; i++)
        {
            j = sortindices [i] ;
            tempb [i] = bl [j+1] ; /* like ineq_row, bl and bu are shifted */
        }
        pproj_copyx (bl+1, tempb, ni) ;
        bu = Prob->bu ;
        for (i = 0; i < ni; i++)
        {
            j = sortindices [i] ;
            tempb [i] = bu [j+1] ; /* like ineq_row, bl and bu are shifted */
        }
        pproj_copyx (bu+1, tempb, ni) ;
    
    /* Evaluate number of columns in each block */
        p = 0 ;
        for (j = 0; j < ncol; j++)
        {
            /* locate the smallest row index, count number elements in row */
            q = Ap [j+1] ;
            k = nrow ;
            for ( ; p < q ; p++)
            {
                i = invperm [Ai [p]] ;
                k = MIN (k, i) ;    /* smallest row index in column */
                rowcount [i]++ ;    /* # elements in row of permuted matrix */
            }
    
            /* count number of columns in block */
            blk_preperm [j] = row_to_blk [k] ;
            ncols [blk_preperm [j]]++ ;
        }
    
        /* set up Kp (pointer to kids), col_start (first column singleton in
           block, col_to_blk (block associated with each column) */
        Kids = W->Kids ;
        Kp = W->Kp ;
        n = m = 0 ;
        for (k = 0; k < blks; k++)
        {
            colperm [k] = Kp [k] = n ;
            n += nkids [k] ;
            l = m + ncols [k] ;
            /* ncols is temporary start of cols in block k */
            ncols [k] = col_start [k] = m ;
            for ( ; m < l ; m++)
            {
                W->col_to_blk [m] = k ;
            }
        }
        col_start [blks] = l ;

        /* Kids is the array of node children pointed into by Kp */
        for (k = 0 ; k < root ; k++)
        {
            Kids [colperm [Cparent [k]]++] = k ;
        }
    
        /* done using Cparent, free it */
        free (Cparent) ;

        /* sol_to_blk [j] = block associated with jth strict inequality */
        sol_to_blk = W->sol_to_blk ;
        sol_to_blk [0] = 0 ; /* not used, put 0 there */
        for (i = 1; i <= ni; i++)
        {
            sol_to_blk [i] = row_to_blk [ineq_row [i]] ;
        }
        sol_to_blk [ni+1] = blks ;

        /* sol_start [k] = first singleton (strict inequality) associated
                           with block k */
        j = 1 ;
        k = sol_to_blk [j] ;
        blk = 0 ;
        /* set sol_start to 1 for all block before first inequality */
        while ( blk < k )
        {
            W->sol_start [blk] = j ;
            blk++ ;
        }

        while ( blk < blks )
        {
            /* k = sol_to_blk [j] and j = first strict inequality in block k */
            W->sol_start [k] = j ;
            j++ ;
            /* find the first row in the next block */
            while ( sol_to_blk [j] <= k )
            {
                j++ ;
            }
            /* store the block associated with the first row */
            k = sol_to_blk [j] ;
            /* set the start of all blocks before this new block to j */
            while ( blk < k )
            {
                blk++ ;
                W->sol_start [blk] = j ;
            }
        }
        W->sol_start [blk] = ni + 1 ;
 
        /* permuted column stored in work arrays */
        worklo = W->lo ;
        workhi = W->hi ;
        worky = W->cold ;
    
        /* colperm [k] = column of original matrix corresponding to column k in
                         permuted matrix */
        for (j = 0; j < ncol; j++)
        {
    
            k = ncols [blk_preperm [j]]++ ;
            colperm [k] = j ;
            worklo [k] = lo [j] ;
            workhi [k] = hi [j] ;
            worky [k] = Prob->y [j] ;
        }
        pproj_copyx (Prob->y, worky, ncol) ;
        pproj_copyx (lo, worklo, ncol) ;
        pproj_copyx (hi, workhi, ncol) ;
    
       /* apply colperm to columns of A and rowperm to rows of A to obtain AT,
          also apply colperm to lo and hi. The goal of the row and
          column permutations is to obtain the structure shown in Figure 3 of
          the multilevel paper (Math. Program., Ser. A (2008) 112:403–425
          DOI 10.1007/s10107-006-0022-3)*/
    
        k = 0 ;
        for (i = 0; i < nrow; i++)
        {
            l = k + rowcount [i] ;
            rowcount [i] = k ;
            ATp  [i] = k ;
            AFTp [i] = k ;
            k = l ;
        }
        ATp  [nrow] = k ;
        AFTp [nrow] = k ;
    
        /* done using blk_preperm above, now use the space to store the
           number of nonzeros in each column of A */
        Anz = blk_preperm ;
        Amax = ZERO ;
        for (col = 0; col < ncol; col++)
        {
            j = colperm [col] ;
            p = Ap [j] ;
            q = Ap [j+1] ;
            Anz [col] = q - p ;
            for (; p < q; p++)
            {
                i = invperm [Ai [p]] ;
                k = rowcount [i]++ ;
                ATi [k] = col ;
                t = Ax [p] ;
                ATx [k] = t ;
                if ( fabs (t) > Amax ) Amax = fabs (t) ;
            }
        }
        if ( (Parm->ScaleSigma == TRUE) && (Amax != ZERO) )
        {
            W->sigma *= Amax ;
        }
    
        /* construct the final A matrix, A = AT' */
        for (j = 0; j < ncol; j++)
        {
            Ap [j+1] = Ap [j] + Anz [j] ;
            Anz [j] = Ap [j] ;
        }
        p = 0 ;
        for (i = 0; i < nrow; i++)
        {
            q = ATp [i+1] ;
            for (; p < q; p++)
            {
                pp = Anz [ATi [p]]++ ;
                Ai [pp] = i ;
                Ax [pp] = ATx [p] ;
            }
        }
        /* A and AT have now been formed */

#ifndef NDEBUG
        /* check that A and AT are correctly permuted */
        n = OrigAp [ncol] ;
        CheckAp = (INT *) malloc ((ncol+1)*sizeof (INT)) ;
        CheckAi = (INT *) malloc (n*sizeof (INT)) ;
        CheckAx = (double *) malloc (n*sizeof (double)) ;
    
        /* copy column pointers and the numerical values to the check matrix */
        pproj_copyI (CheckAp, Ap, ncol+1) ;
        pproj_copyx (CheckAx, Ax, n) ;
    
        /* map rows of the current permuted A to rows of the original matrix*/
        for (p = 0; p < n; p++)
        {
            CheckAi [p] = rowperm [Ai [p]] ;
        }
    
        /* now transpose CheckA */
        CheckATp = (INT *) malloc ((nrow+1)*sizeof (INT)) ;
        CheckATi = (INT *) malloc (n*sizeof (INT)) ;
        CheckATx = (double *) malloc (n*sizeof (double)) ;
        pproj_transpose (CheckATp, CheckATi, CheckATx, CheckAp, CheckAi,
                         CheckAx, nrow, ncol, tworkstart) ;
    
        /* map the columns in checkA to those in the original matrix */
        for (p = 0; p < n; p++)
        {
            CheckATi [p] = colperm [CheckATi [p]] ;
        }
        pproj_transpose (CheckAp, CheckAi, CheckAx, CheckATp, CheckATi,
                         CheckATx, ncol, nrow, tworkstart) ;
    
        /* checkA  should now equal the original matrix */
        for (j = 0; j <= ncol; j++)
        {
            if ( OrigAp [j] != CheckAp [j] )
            {
                pproj_error (-1, __FILE__, __LINE__,
                "error in Ap in initlevels\n") ;
            }
        }
        /* check Ai and Ax */
        for (p = 0; p < n; p++)
        {
            if ( OrigAi [p] != CheckAi [p] )
            {
                printf ("p: %i Ai: %i check: %i\n", p, Ai [p], CheckAi [p]) ;
                pproj_error (-1, __FILE__, __LINE__,
                "error in Ai in initlevels\n") ;
            }
            if ( OrigAx [p] != CheckAx [p] )
            {
                printf ("p: %i Ax: %e check: %e\n", p, Ax [p], CheckAx [p]) ;
                pproj_error (-1, __FILE__, __LINE__,
                "error in Ax in initlevels\n") ;
            }
        }
        /* check transpose */
        pproj_transpose (CheckATp, CheckATi, CheckATx, Ap, Ai, Ax,
                         nrow, ncol, tworkstart) ;
        /* check ATp */
        for (i = 0; i <= nrow; i++)
        {
            if ( ATp [i] != CheckATp [i] )
            {
                pproj_error (-1, __FILE__, __LINE__,
                "error in ATp in initlevels\n") ;
            }
        }
        /* check ATi and ATx */
        for (p = 0; p < n; p++)
        {
            if ( ATi [p] != CheckATi [p] )
            {
                printf ("p: %i ATi: %i check: %i\n", p, ATi [p], CheckATi [p]) ;
                pproj_error (-1, __FILE__, __LINE__,
                "error in ATi in initlevels\n") ;
            }
            if ( ATx [p] != CheckATx [p] )
            {
                printf ("p: %i ATx: %e check: %e\n", p, ATx [p], CheckATx [p]) ;
                pproj_error (-1, __FILE__, __LINE__,
                "error in ATx in initlevels\n") ;
            }
        }
        free (CheckAp) ;
        free (CheckATp) ;
        free (OrigAp) ;
        free (CheckAi) ;
        free (CheckATi) ;
        free (OrigAi) ;
        free (CheckAx) ;
        free (CheckATx) ;
        free (OrigAx) ;
        free (invperm) ;
        free (rowcount) ;
        free (blk_preperm) ;
        free (ncols) ;
        free (row_to_blk) ;
#endif
        free (tworkstart) ;
    }
    else /* for an iterative method, the row permutation is the identity; we
            just need to compute AT and the column pointers for AFT */
    {
        blks = 1 ;
        if ( Parm->ScaleSigma == TRUE )
        {
            Amax = pproj_max (Ax, Annz) ;
            if ( Amax != ZERO )
            {
                W->sigma *= Amax ;
            }
        }
        pproj_transpose (ATp, ATi, ATx, Ap, Ai, Ax, nrow, ncol,
                         W->AFTp /* work array */) ;
        /* store pointers in AFTp */
        pproj_copyI (AFTp, ATp, nrow+1) ;
        /* copy Prob->b, Prob->lo, and Prob->hi to work structure */
        pproj_copyx (W->b, Prob->b, nrow) ;
        pproj_copyx (W->lo, Prob->lo, ncol) ;
        pproj_copyx (W->hi, Prob->hi, ncol) ;
        if ( Parm->use_lambda == TRUE )
        {
            pproj_copyx (W->shift_l, W->lambda, nrow) ;
        }
        else
        {
            pproj_initx (W->shift_l, ZERO, nrow) ;
        }
        pproj_initx (W->lambda, ZERO, nrow) ;
        Stat->solves = (INT *) malloc (sizeof (INT)) ;
        Stat->solves [0] = 0 ;
        W->sol_to_blk = NULL ;
        W->ustart = NULL ;
        W->lstart = NULL ;

#ifndef NDEBUG
        W->leftdesc  = (int *) malloc (blks*sizeof (int)) ;
        W->col_start = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        W->row_start = (INT *) malloc ((blks+1)*sizeof (INT)) ;
        W->leftdesc [0] = 0 ;
        W->row_start [0] = 0 ;
        W->row_start [1] = nrow ;
        W->col_start [0] = 0 ;
        W->col_start [1] = ncol ;
#endif
    }
    W->blks = blks ;
    Stat->blks = blks ;
    /* Both AT and A are now permuted according to the rowperm and colperm */

    /* Complete the allocation of the workspace W by allocating the
       pointers to the starting column in each block of the multilevel
       decomposition. lstart is associated with the equations at lower
       bounds and ustart is associated with the equations at upper bound */
    W->lstart = (INT *) malloc ((blks+1)*sizeof (INT)) ;
    W->ustart = (INT *) malloc ((blks+1)*sizeof (INT)) ;

    /* determine the length of the longest row in A */
    maxrow = 0 ;
    p = 0 ;
    for (i = 1; i <= nrow; i++) 
    {
        q = ATp [i] ;
        if ( q - p > maxrow ) maxrow = q - p ;
        p = q ;
    }
    Prob->maxrow = maxrow ;

    ki = 0 ; /* INT space */
    kd = 0 ; /* double space */
    ntot = ni + ncol + 1 ;

    /* space for phase1 */
    if ( Parm->phase1 >= 0 )
    {
        j = 3*maxrow ;
        if ( j > ki ) ki = j ;

        j = nrow + ni + 1 ;
        if ( j > ki ) ki = j ;

        j = 2*nrow + ncol + ntot ;
        if ( j > kd ) kd = j ;
    }

    /* space for SpaRSA */
    if ( Parm->use_sparsa == TRUE )
    {
        if ( sizeof (INT) >= sizeof (int) )
        {
            j = ntot + nrow + ncol ;
            if ( j > ki ) ki = j ;
        }
        else
        {
            i = ceil ((double) sizeof (int) / (double) sizeof (INT)) ;
            j = ntot + nrow + i*ncol ;
            if ( j > ki ) ki = j ;
        }

        j = ncol + ntot + nrow + Parm->mem ;
        if ( j > kd ) kd = j ;
    }

    /* space for coordinate ascent steps */
    if ( Parm->use_coor_ascent == TRUE )
    {
        j = 3*maxrow ;
        if ( j > ki ) ki = j ; /* INT space */

        j = maxrow ;
        if ( j > kd ) kd = j ; /* double space */
    }

    /* Allocate space for ssor algorithms. These are either used in the
       iterative mode or in a presolve phase before starting the updates
       and downdates */

    if ( ((Parm->cholmod == TRUE) && (Parm->use_ssor0 == TRUE)) ||
          (Parm->cholmod == FALSE) )
    {
        if ( ntot + ncol > ki )
        {
            ki = ntot + ncol ; /* INT space */
        }

        j = 2*nrow +ncol + ntot ;
        if ( j > kd ) kd = j ; /* double space */
    }

    if ( ((Parm->cholmod == TRUE) && (Parm->use_ssor1 == TRUE)) ||
         (Parm->cholmod == FALSE) )
    {
        if ( ntot + ncol + nrow > ki )
        {
            ki = ntot + ncol + nrow ; /* INT space */
        }
        j = 5*nrow + ncol + ntot ;
        k = MIN (nrow, Parm->ssormem) ;
        k = MIN (k, ncol) ;
        k = MAX (k, 2) ;         /* ssormem must be at least 2 */
        W->ssormem = k ;
        j += (4+nrow)*k + (k*(k-1))/2 ;
        if ( j > kd ) kd = j ;   /* double space */
    }

    /* initialize some statistics */
    W->nrup = 0 ;            /* total number of rooted updates */
    W->npup = 0 ;            /* total number of partial updates */
    W->nchols = 0 ;          /* total number of chols */
    W->cholaatflops = ZERO ; /* average flops in a chol (incl. AA') */
    W->rupflops = ZERO ;     /* average flops in a rooted update */
    W->pupflops = ZERO ;     /* average flops in a partial update */

    if ( Parm->cholmod == TRUE ) /* use update/downdate */
    {
        /* -> line search */
        j = (ntot + 1)*sizeof(INT) ;
        if ( ki < j ) ki = j ;

        j = (2*ncol + 3*nrow + ni)*sizeof(double) ;
        if ( kd < j ) kd = j ;
        /* (2*ncol + 3*nrow + ni)*sizeof(double) + (ntot + 1)*sizeof(INT)) */

        /* -> update/downdate at top of ascent */
        j = (2*ncol + 4*nrow)*sizeof(INT) ;
        if ( ki < j ) ki = j ;

        j = (MAX (nrow, ncol) + nrow)*sizeof(double) ;
        if ( kd < j) kd = j ;
        /* (MAX (nrow, ncol) + nrow)*sizeof(double)
                 + (2*ncol + 4*nrow)*sizeof(INT)) */

        /* -> update_Anz */
        j = (ncol+3*nrow)*sizeof(INT) ;
        if ( ki < j ) ki = j ;

        j = nrow*sizeof(double) ;
        if ( kd < j ) kd = j ;
        /* 2*nrow*sizeof(double) + (ncol+3*nrow)*sizeof(INT) */

        /* -> modrow inside big iteration updates */
        j = nrow*sizeof(INT) ;
        if ( ki < j ) ki = j ;

        j = 4*nrow*sizeof(double) ;
        if ( kd < j ) kd = j ;
        /* 4*nrow*sizeof(double) + nrow*sizeof(INT) */

        /* -> modrow at start of big iteration */
        j = (nrow+ncol)*sizeof(INT) ;
        if ( ki < j ) ki = j ;

        j = (2*nrow+ncol)*sizeof(double) ;
        if ( kd < j ) kd = j ;
        /* (2*nrow+ncol)*sizeof(double) + (nrow+ncol)*sizeof(INT) */

        /* -> error check */
        j = ntot*sizeof(INT) ;
        if ( ki < j ) ki = j ;

        j = (3*nrow+ntot)*sizeof(double) ;
        if ( kd < j ) kd = j ;
        /* (3*nrow+ntot)*sizeof(double) + ntot*sizeof(INT) */

        /* -> line search */
        j = (ntot+1)*sizeof(INT) ;
        if ( ki < j ) ki = j ;

        j = (3*nrow+2*ncol+ni)*sizeof(double) ;
        if ( kd < j ) kd = j ;
        /* (3*nrow+2*ncol+ni)*sizeof(double) + (ntot+1)*sizeof(INT) */

        /* -> multilevel implementation */
        if ( Parm->multilevel )
        {
            j = (4*nrow+3*ncol)*sizeof (INT) ;
            if ( ki < j ) ki = j ;
        }
    }

    /* allocate INT work space */
    W->arrayi = (INT *) malloc (ki*sizeof (INT)) ;
    if ( W->arrayi == NULL )
    {
        return (PPROJ_OUT_OF_MEMORY) ;
    }

    /* allocate double work space */
    W->arrayd = (double *) malloc (kd*sizeof (double)) ;
    if ( W->arrayd == NULL )
    {
        return (PPROJ_OUT_OF_MEMORY) ;
    }

    return (PPROJ_STATUS_OK) ;
}
