/* =========================================================================
   ============================== pproj_initwork ===========================
   =========================================================================
    Initialize the working arrays used to implement the algorithm
   ========================================================================== */

#include "pproj.h"

int pproj_initwork
(
    PPcom *I
)
{
    int size_updowns ;
    INT  anz, k, nrow, ncol, ni ;
    PPparm *Parm ;
    PPprob *Prob ;
    PPwork    *W ;
    PPstat *Stat ;

    Parm = I->Parm ;
    W = I->Work ;
    Prob = I->Prob ;
    Stat = I->Stat ;

    nrow = Prob->nrow ;
    ncol = Prob->ncol ;
    ni = Prob->ni ;
    anz = Prob->Ap [ncol] ;
    /* size of the ns array */
    k = ncol + ni + 1 ;
    k = MAX (k, nrow) ;
    /* size of the updowns array in the Stat structure */
    size_updowns = (int) 3.*sqrt ((double) ncol) ;
    size_updowns = MAX (size_updowns, 10) ;
    Stat->size_updowns = size_updowns ;

    /* initially the matrix is not factorized */
    W->fac = FALSE ;
    W->L = NULL ;

    /* initial value for the proximal regularization parameter */
    W->sigma = Parm->sigma ;

    /* initial value for SpaRSA stopping parameter */
    W->gamma = Parm->gamma ;

    /* initialize the change variables to be 1 that that either
       coordinate ascent or ssor iterations are performed if possible */
    W->chg_coor = 1 ;
    W->chg_ssor0 = 1 ;
    W->chg_ssor1 = 1 ;
    W->chg_sparsa = 1 ;
    W->sparsaOK = TRUE ;

    if ( !Parm->providePPcom )
    {
        W->ATp = (INT *) malloc ((nrow+1)*sizeof (INT)) ;
        W->ATi = (INT *) malloc (anz*sizeof (INT)) ;
        W->ATx = (double *) malloc (anz*sizeof (double)) ;
    
        W->AFTp = (INT *) malloc ((nrow+1)*sizeof (INT)) ;
        W->AFTnz = (INT *) malloc (nrow*sizeof (INT)) ;
        W->AFTi = (INT *) malloc (anz*sizeof (INT)) ;
        W->AFTx = (double *) malloc (anz*sizeof (double)) ;
    
        /* lambda + shift_l */
        W->lambda_tot = (double *) malloc (nrow*sizeof (double)) ;
    
        /* the change in lambda */
        W->dlambda = (double *) malloc (nrow*sizeof (double)) ;
    
        /* ns used in line search, points from index to break point number.
           Also, used for bookkeeping in check_error */
        W->ns = (INT *) malloc (k*sizeof (INT)) ;
    
        /* empty if row not modified, otherwise points into Rowmodlist
           which contains list of rows to modify*/
        W->RowmodFlag = (INT *) malloc ((ni+1)*sizeof(INT)) ;
    
        /* first part of this list contains the rows to delete
           the tail contains the rows to add */
        W->RowmodList = (INT *) malloc (ni*sizeof(INT)) ;
    
        /* empty if column not modified, otherwise points into Colmodlist
           which contains list of columns to modify*/
        W->ColmodFlag = (INT *) malloc (ncol*sizeof(INT)) ;
    
        /* first part of this list contains the columns to add
           the tail contains the columns to delete */
        W->ColmodList = (INT *) malloc (ncol*sizeof(INT)) ;
    
        /* list of rows that are dropped since last call to updateAnz */
        W->dropped = malloc (ni*sizeof (INT)) ;
    
        /* list of free variables */
        W->F = (INT *) malloc (ncol*sizeof (INT)) ;
    
        /* ib [col] = +1 if column at upper bound
                      -1 if column at lower bound
                       0 if column is free */
        W->ib = (INT *) malloc (ncol*sizeof (INT)) ;
    
        /* proximal lambda */
        W->shift_l = (double *) malloc (nrow*sizeof (double)) ;
     
        W->b = (double *) malloc (nrow*sizeof (double)) ;
        W->c = (double *) malloc (ncol*sizeof (double)) ;
        W->cold = (double *) malloc (ncol*sizeof (double)) ;
    
        W->lo = (double *) malloc (ncol*sizeof (double)) ;
        W->hi = (double *) malloc (ncol*sizeof (double)) ;
        W->D = (double *) malloc (nrow*sizeof(double)) ;
    
        /* links for active rows */
        W->RLinkUp = (INT *) malloc ((nrow+1)*sizeof (INT)) ;
        W->RLinkDn = (INT *) malloc ((nrow+1)*sizeof (INT)) ;
    
        /* Links for active singletons. Two links are stored in the same array.
           uLink points to the strict inequalities at upper bound and
           lLink points to the strict inequalities at lower bound.
           The locations of the strict inequality rows start at 1, so the
           last element is ni. ni+1 marks the end of lLink and ni+2 marks the
           end of uLink */
        W->SLinkUp = (INT *) malloc ((ni+3)*sizeof (INT)) ;
        W->SLinkDn = (INT *) malloc ((ni+3)*sizeof (INT)) ;
    
        /* ir [row] = 0         for an equality constraint
                    =  strict # for an active inequality at upper bound
                    = -strict # for an active inequality at lower bound
                    =  strict # + ni for a dropped constraint */
        W->ir = (INT *) malloc (nrow*sizeof(INT)) ;
        /* updowns counts the number of updates of each size */
        Stat->updowns = (int *) malloc ((size_updowns+1)*sizeof (int)) ;
        if ( Stat->updowns == NULL )
        {
            return (PPROJ_OUT_OF_MEMORY) ;
        }
        if ( Parm->cholmod )
        {
            /* changeRHS stores the right side change. It is initialized to be
               zero and then restored to 0 after being used to update the change
               in the forward solve due to the change in the right side */
            W->changeRHS = (double *) malloc (nrow*sizeof (double)) ;

            /* The column in A that are updated are pointed to by Cp while Cnz
               gives the number of nonzeros in the column */
            W->Cp = (INT *) malloc (ncol*sizeof (INT)) ;
            W->Cnz = (INT *) malloc (ncol*sizeof (INT)) ;

            /* Store a copy of A in the work structure as a packed cholmod
               sparse matrix with sorted columns. Note that W->A stores
               pointers to the matrix in the Prob structure */
            W->A = (cholmod_sparse *) malloc (sizeof (cholmod_sparse)) ;
            pproj_cholmod_sparse (W->A, nrow, ncol, Prob->Ap [ncol], Prob->Ap,
                       Prob->Anz, Prob->Ai, Prob->Ax, TRUE, TRUE, CHOLMOD_REAL);
            if ( (W->A == NULL) || (W->Cnz == NULL) )
            {
                return (PPROJ_OUT_OF_MEMORY) ;
            }
        }
    }

    /* initialize some of the variables */
    pproj_initI (W->RowmodFlag, (INT) EMPTY, ni+1) ;
    pproj_initI (W->ColmodFlag, (INT) EMPTY, ncol) ;
    pproj_initI (W->ns, (INT) EMPTY, k) ;
    if ( Parm->use_lambda )
    {
        pproj_copyx (W->shift_l, W->lambda, nrow) ;
    }

    W->nrowadd = 0 ;
    W->nrowdel = 0 ;
    W->ncoladd = 0 ;
    W->ncoldel = 0 ;

    /* Initialization for update/down */
    if ( Parm->cholmod )
    {
        /* number of dropped rows */
        W->nd = 0 ;
        pproj_initx (W->changeRHS, ZERO, nrow) ;
    }

#ifndef NDEBUG
    pproj_initx (W->dlambda, ZERO, nrow) ;
    pproj_initx (W->x, ZERO, ncol) ;
#endif
    /* remainder of workspace malloc'ed in lpdasa_initlevels */

    /* initialize counting variables in Stat structure */

    Stat->solves        = NULL ; /* Stat->solves initialized in initlevels */
    Stat->phase1_its       = 0 ; /* # phase1 iterations */
    Stat->coor_ascent_its  = 0 ; /* # coordinate ascent iterations */
    Stat->ssor0_its        = 0 ; /* # ssor0 iterations */
    Stat->ssor1_its        = 0 ; /* # ssor1 iterations */
    Stat->sparsa_its       = 0 ; /* # sparsa iterations */
    Stat->nprox            = 0 ; /* # proximal updates */
    Stat->coldn            = 0 ; /* # column downdates to L */
    Stat->colup            = 0 ; /* # column updates to L */
    Stat->rowdn            = 0 ; /* # rows dropped from L */
    Stat->rowup            = 0 ; /* # rows added to L */
    Stat->coor_ascent_free = 0 ; /* # of variables freed in coordinate ascent */
    Stat->coor_ascent_drop = 0 ; /* # of rows dropped in coordinate ascent */
    Stat->ssor0_free       = 0 ; /* # of variables freed in ssor0 */
    Stat->ssor0_drop       = 0 ; /* # of rows dropped in ssor0 */
    Stat->ssor1_free       = 0 ; /* # of variables freed in ssor1 */
    Stat->ssor1_drop       = 0 ; /* # of rows dropped in ssor1 */
    Stat->sparsa_col       = 0 ; /* # of changes bound constraints in sparsa */
    Stat->sparsa_row       = 0 ; /* # of changes row constraints in sparsa */
    Stat->sparsa_step_fail = 0 ; /* # of failures of Armijo step in sparsa */
    Stat->nchols           = 0 ; /* # of Cholesky factorizations */
    Stat->lnnz             = 0 ; /* # of nonzeros in Cholesky factor */
    Stat->maxdepth         = 0 ; /* depth of multilevel partition tree */

    /* initialize timers in Stat structure */
    Stat->partition   = ZERO ; /* compute reordering of rows of A */
    Stat->initialize  = ZERO ; /* initwork and initlevels, includes partition */
    Stat->phase1      = ZERO ; /* time in phase1 */
    Stat->sparsa      = ZERO ; /* sparsa */
    Stat->coor_ascent = ZERO ; /* coordinate ascent */
    Stat->ssor0       = ZERO ; /* ssor0 */
    Stat->ssor1       = ZERO ; /* ssor1 */
    Stat->dasa        = ZERO ; /* dasa (includes coor_ascent, ssor0, ssor1) */
    Stat->dasa_line   = ZERO ; /* dasa line search */
    Stat->checkerr    = ZERO ; /* check_error */
    Stat->prox_update = ZERO ; /* proximal update */
    Stat->invert      = ZERO ; /* invert permutation of rows and columns */
    Stat->modrow      = ZERO ; /* modrow (update L by adding or deleting rows)*/
    Stat->modcol      = ZERO ; /* modcol (rank 1 column updates of L) */
    Stat->chol        = ZERO ; /* cholmod_analyze, cholmod_factorize */
    Stat->cholinc     = ZERO ; /* incremental cholmod_rowfac */
    Stat->dltsolve    = ZERO ; /* dltsolve (back solve) */
    Stat->lsolve      = ZERO ; /* lsolve (forward solve) */
    Stat->nrow = nrow ;        /* needed to estimate Cholesky factor sparsity */
    /* initialize number of updates and downdates of each size to 0 */
    pproj_initi (Stat->updowns, 0, (size_updowns+1)) ;
    return (PPROJ_STATUS_OK) ;
}
